import os
from tempfile import TemporaryDirectory
from pathlib import Path

import numpy as np

import pystencils as ps
from pystencils import create_data_handling, create_kernel
from pystencils.datahandling.pycuda import PyCudaArrayHandler
from pystencils.enums import Target

try:
    import pytest
except ImportError:
    import unittest.mock
    pytest = unittest.mock.MagicMock()

SCRIPT_FOLDER = Path(__file__).parent.absolute()
INPUT_FOLDER = SCRIPT_FOLDER / "test_data"


def basic_iteration(dh):
    dh.add_array('basic_iter_test_gl_default')
    dh.add_array('basic_iter_test_gl_3', ghost_layers=3)

    for b in dh.iterate():
        assert b.shape == b['basic_iter_test_gl_3'].shape
        assert b.shape == b['basic_iter_test_gl_default'].shape


def access_and_gather(dh, domain_size):
    dh.add_array('f1', dtype=np.dtype(np.int8))
    dh.add_array_like('f2', 'f1')
    dh.add_array('v1', values_per_cell=3, dtype=np.int64, ghost_layers=2)
    dh.add_array_like('v2', 'v1')

    dh.swap('f1', 'f2')
    dh.swap('v1', 'v2')

    # Check symbolic field properties
    assert dh.fields.f1.index_dimensions == 0
    assert dh.fields.f1.spatial_dimensions == len(domain_size)
    assert dh.fields.f1.dtype.numpy_dtype == np.int8

    assert dh.fields.v1.index_dimensions == 1
    assert dh.fields.v1.spatial_dimensions == len(domain_size)
    assert dh.fields.v1.dtype.numpy_dtype == np.int64

    for b in dh.iterate(ghost_layers=0):
        val = sum(b.cell_index_arrays)
        np.copyto(b['f1'], val)
        for i, coord_arr in enumerate(b.cell_index_arrays):
            np.copyto(b['v1'][..., i], coord_arr)

    full_arr = dh.gather_array('v1')
    if full_arr is not None:
        expected_shape = domain_size + (3,)
        assert full_arr.shape == expected_shape
        for x in range(full_arr.shape[0]):
            for y in range(full_arr.shape[1]):
                if len(domain_size) == 3:
                    for z in range(full_arr.shape[2]):
                        assert full_arr[x, y, z, 0] == x
                        assert full_arr[x, y, z, 1] == y
                        assert full_arr[x, y, z, 2] == z
                else:
                    assert len(domain_size) == 2
                    assert full_arr[x, y, 0] == x
                    assert full_arr[x, y, 1] == y

    full_arr = dh.gather_array('f1')
    if full_arr is not None:
        expected_shape = domain_size
        assert full_arr.shape == expected_shape
        for x in range(full_arr.shape[0]):
            for y in range(full_arr.shape[1]):
                if len(domain_size) == 3:
                    for z in range(full_arr.shape[2]):
                        assert full_arr[x, y, z] == x + y + z
                else:
                    assert len(domain_size) == 2
                    assert full_arr[x, y] == x + y


def synchronization(dh, test_gpu=False):
    field_name = 'comm_field_test'
    if test_gpu:
        try:
            from pycuda import driver
            import pycuda.autoinit
        except ImportError:
            return
        field_name += 'Gpu'

    dh.add_array(field_name, ghost_layers=1, dtype=np.int8, cpu=True, gpu=test_gpu)

    # initialize everything with 1
    for b in dh.iterate(ghost_layers=1):
        b[field_name].fill(1)
    for b in dh.iterate(ghost_layers=0):
        b[field_name].fill(42)

    if test_gpu:
        dh.to_gpu(field_name)
        dh.synchronization_function_gpu(field_name)()
    else:
        dh.synchronization_function_cpu(field_name)()


    if test_gpu:
        dh.to_cpu(field_name)

    for b in dh.iterate(ghost_layers=1):
        np.testing.assert_equal(42, b[field_name])


def kernel_execution_jacobi(dh, target):

    test_gpu = target == Target.GPU
    dh.add_array('f', gpu=test_gpu)
    dh.add_array('tmp', gpu=test_gpu)

    if test_gpu:
        assert dh.is_on_gpu('f')
        assert dh.is_on_gpu('tmp')

    stencil_2d = [(1, 0), (-1, 0), (0, 1), (0, -1)]
    stencil_3d = [(1, 0, 0), (-1, 0, 0), (0, 1, 0), (0, -1, 0), (0, 0, 1), (0, 0, -1)]
    stencil = stencil_2d if dh.dim == 2 else stencil_3d

    @ps.kernel
    def jacobi():
        dh.fields.tmp.center @= sum(dh.fields.f.neighbors(stencil)) / len(stencil)

    kernel = create_kernel(jacobi, config=ps.CreateKernelConfig(target=target)).compile()
    for b in dh.iterate(ghost_layers=1):
        b['f'].fill(42)
    dh.run_kernel(kernel)
    for b in dh.iterate(ghost_layers=0):
        np.testing.assert_equal(b['f'], 42)


def vtk_output(dh):
    pytest.importorskip('pyevtk')
    dh.add_array('scalar_field')
    dh.add_array('vector_field', values_per_cell=dh.dim)
    dh.add_array('multiple_scalar_field', values_per_cell=9)
    dh.add_array('flag_field', dtype=np.uint16)

    fields_names = ['scalar_field', 'vector_field', 'multiple_scalar_field', 'flag_field']
    with TemporaryDirectory() as tmp_dir:
        writer1 = dh.create_vtk_writer(os.path.join(tmp_dir, "out1"), fields_names, ghost_layers=True)
        writer2 = dh.create_vtk_writer(os.path.join(tmp_dir, "out2"), fields_names, ghost_layers=False)
        masks_to_name = {1: 'flag1', 5: 'some_mask'}
        writer3 = dh.create_vtk_writer_for_flag_array(os.path.join(tmp_dir, "out3"), 'flag_field', masks_to_name)
        writer1(1)
        writer2(1)
        writer3(1)


def reduction(dh):
    float_seq = [1.0, 2.0, 3.0]
    int_seq = [1, 2, 3]
    for op in ('min', 'max', 'sum'):
        assert (dh.reduce_float_sequence(float_seq, op) == float_seq).all()
        assert (dh.reduce_int_sequence(int_seq, op) == int_seq).all()


def test_symbolic_fields():
    dh = create_data_handling(domain_size=(5, 7))
    dh.add_array('f1', values_per_cell=dh.dim)
    assert dh.fields['f1'].spatial_dimensions == dh.dim
    assert dh.fields['f1'].index_dimensions == 1

    dh.add_array_like("f_tmp", "f1", latex_name=r"f_{tmp}")
    assert dh.fields['f_tmp'].spatial_dimensions == dh.dim
    assert dh.fields['f_tmp'].index_dimensions == 1

    dh.swap('f1', 'f_tmp')


def test_access():
    for domain_shape in [(2, 3, 4), (2, 4)]:
        for f_size in (1, 4):
            dh = create_data_handling(domain_size=domain_shape)
            dh.add_array('f1', values_per_cell=f_size)
            assert dh.dim == len(domain_shape)

            for b in dh.iterate(ghost_layers=1):
                if f_size > 1:
                    assert b['f1'].shape == tuple(ds+2 for ds in domain_shape) + (f_size,)
                else:
                    assert b['f1'].shape == tuple(ds + 2 for ds in domain_shape)

            for b in dh.iterate(ghost_layers=0):
                if f_size > 1:
                    assert b['f1'].shape == domain_shape + (f_size,)
                else:
                    assert b['f1'].shape == domain_shape


def test_access_and_gather():
    for domain_shape in [(2, 2, 3), (2, 3)]:
        dh = create_data_handling(domain_size=domain_shape, periodicity=True)
        access_and_gather(dh, domain_shape)
        synchronization(dh, test_gpu=False)
        synchronization(dh, test_gpu=True)


def test_kernel():
    for domain_shape in [(4, 5), (3, 4, 5)]:
        dh = create_data_handling(domain_size=domain_shape, periodicity=True)
        assert all(dh.periodicity)
        kernel_execution_jacobi(dh, Target.CPU)
        reduction(dh)

        try:
            import pycuda
            dh = create_data_handling(domain_size=domain_shape, periodicity=True)
            kernel_execution_jacobi(dh, Target.GPU)
        except ImportError:
            pass


@pytest.mark.parametrize('target', (Target.CPU, Target.GPU))
def test_kernel_param(target):
    for domain_shape in [(4, 5), (3, 4, 5)]:
        if target == Target.GPU:
            pytest.importorskip('pycuda')

        dh = create_data_handling(domain_size=domain_shape, periodicity=True, default_target=target)
        kernel_execution_jacobi(dh, target)
        reduction(dh)


def test_vtk_output():
    pytest.importorskip('pyevtk')
    for domain_shape in [(4, 5), (3, 4, 5)]:
        dh = create_data_handling(domain_size=domain_shape, periodicity=True)
        vtk_output(dh)


def test_add_arrays():
    domain_shape = (3, 4, 5)
    field_description = 'x, y(9)'

    dh = create_data_handling(domain_size=domain_shape, default_ghost_layers=0, default_layout='numpy')
    x_, y_ = dh.add_arrays(field_description)

    x, y = ps.fields(field_description + ': [3,4,5]')

    assert x_ == x
    assert y_ == y
    assert x == dh.fields['x']
    assert y == dh.fields['y']


def test_get_kwarg():
    domain_shape = (10, 10)
    field_description = 'src, dst'

    dh = create_data_handling(domain_size=domain_shape, default_ghost_layers=1)
    src, dst = dh.add_arrays(field_description)
    dh.fill("src", 1.0, ghost_layers=True)
    dh.fill("dst", 0.0, ghost_layers=True)

    with pytest.raises(ValueError):
        dh.add_array('src')

    ur = ps.Assignment(src.center, dst.center)
    kernel = ps.create_kernel(ur).compile()

    kw = dh.get_kernel_kwargs(kernel)
    assert np.all(kw[0]['src'] == dh.cpu_arrays['src'])
    assert np.all(kw[0]['dst'] == dh.cpu_arrays['dst'])


def test_add_custom_data():
    pytest.importorskip('pycuda')

    import pycuda.gpuarray as gpuarray
    import pycuda.autoinit  # noqa

    def cpu_data_create_func():
        return np.ones((2, 2), dtype=np.float64)

    def gpu_data_create_func():
        return gpuarray.zeros((2, 2), dtype=np.float64)

    def cpu_to_gpu_transfer_func(gpuarr, cpuarray):
        gpuarr.set(cpuarray)

    def gpu_to_cpu_transfer_func(gpuarr, cpuarray):
        gpuarr.get(cpuarray)

    dh = create_data_handling(domain_size=(10, 10))
    dh.add_custom_data('custom_data',
                       cpu_data_create_func,
                       gpu_data_create_func,
                       cpu_to_gpu_transfer_func,
                       gpu_to_cpu_transfer_func)

    assert np.all(dh.custom_data_cpu['custom_data'] == 1)
    assert np.all(dh.custom_data_gpu['custom_data'].get() == 0)

    dh.to_cpu(name='custom_data')
    dh.to_gpu(name='custom_data')

    assert 'custom_data' in dh.custom_data_names


def test_log():
    dh = create_data_handling(domain_size=(10, 10))
    dh.log_on_root()
    assert dh.is_root
    assert dh.world_rank == 0


def test_save_data():
    domain_shape = (2, 2)

    dh = create_data_handling(domain_size=domain_shape, default_ghost_layers=1)
    dh.add_array("src", values_per_cell=9)
    dh.fill("src", 1.0, ghost_layers=True)
    dh.add_array("dst", values_per_cell=9)
    dh.fill("dst", 1.0, ghost_layers=True)

    dh.save_all(str(INPUT_FOLDER) + '/datahandling_save_test')


def test_load_data():
    domain_shape = (2, 2)

    dh = create_data_handling(domain_size=domain_shape, default_ghost_layers=1)
    dh.add_array("src", values_per_cell=9)
    dh.fill("src", 0.0, ghost_layers=True)
    dh.add_array("dst", values_per_cell=9)
    dh.fill("dst", 0.0, ghost_layers=True)

    dh.load_all(str(INPUT_FOLDER) + '/datahandling_load_test')
    assert np.all(dh.cpu_arrays['src']) == 1
    assert np.all(dh.cpu_arrays['dst']) == 1

    domain_shape = (3, 3)

    dh = create_data_handling(domain_size=domain_shape, default_ghost_layers=1)
    dh.add_array("src", values_per_cell=9)
    dh.fill("src", 0.0, ghost_layers=True)
    dh.add_array("dst", values_per_cell=9)
    dh.fill("dst", 0.0, ghost_layers=True)
    dh.add_array("dst2", values_per_cell=9)
    dh.fill("dst2", 0.0, ghost_layers=True)

    dh.load_all(str(INPUT_FOLDER) + '/datahandling_load_test')
    assert np.all(dh.cpu_arrays['src']) == 0
    assert np.all(dh.cpu_arrays['dst']) == 0
    assert np.all(dh.cpu_arrays['dst2']) == 0


def test_array_handler():
    size = (2, 2)
    pytest.importorskip('pycuda')
    array_handler = PyCudaArrayHandler()

    zero_array = array_handler.zeros(size)
    cpu_array = np.empty(size)
    array_handler.download(zero_array, cpu_array)
    assert np.all(cpu_array) == 0

    ones_array = array_handler.ones(size)
    cpu_array = np.empty(size)
    array_handler.download(ones_array, cpu_array)
    assert np.all(cpu_array) == 1

    empty = array_handler.empty(size)
    assert empty.strides == (16, 8)
    empty = array_handler.empty(shape=size, layout=(1, 0))
    assert empty.strides == (8, 16)

    random_array = array_handler.randn(size)




"""Tests for the (un)packing (from)to buffers on a CUDA GPU."""

import numpy as np
import pytest

import pystencils
from pystencils import Assignment, Field, FieldType, CreateKernelConfig, create_kernel
from pystencils.field import create_numpy_array_with_layout, layout_string_to_tuple
from pystencils.slicing import (
    add_ghost_layers, get_ghost_region_slice, get_slice_before_ghost_layer)
from pystencils.stencil import direction_string_to_offset

try:
    # noinspection PyUnresolvedReferences
    import pycuda.autoinit
    import pycuda.gpuarray as gpuarray
except ImportError:
    pass


FIELD_SIZES = [(4, 3), (9, 3, 7)]


def _generate_fields(dt=np.uint8, stencil_directions=1, layout='numpy'):
    pytest.importorskip('pycuda')
    field_sizes = FIELD_SIZES
    if stencil_directions > 1:
        field_sizes = [s + (stencil_directions,) for s in field_sizes]

    fields = []
    for size in field_sizes:
        field_layout = layout_string_to_tuple(layout, len(size))
        src_arr = create_numpy_array_with_layout(size, field_layout).astype(dt)

        array_data = np.reshape(np.arange(1, int(np.prod(size)+1)), size)
        # Use flat iterator to input data into the array
        src_arr.flat = add_ghost_layers(array_data,
                                        index_dimensions=1 if stencil_directions > 1 else 0).astype(dt).flat

        gpu_src_arr = gpuarray.to_gpu(src_arr)
        gpu_dst_arr = gpuarray.empty_like(gpu_src_arr)
        size = int(np.prod(src_arr.shape))
        gpu_buffer_arr = gpuarray.zeros(size, dtype=dt)

        fields.append((src_arr, gpu_src_arr, gpu_dst_arr, gpu_buffer_arr))
    return fields


def test_full_scalar_field():
    """Tests fully (un)packing a scalar field (from)to a GPU buffer."""
    fields = _generate_fields()
    for (src_arr, gpu_src_arr, gpu_dst_arr, gpu_buffer_arr) in fields:
        src_field = Field.create_from_numpy_array("src_field", src_arr)
        dst_field = Field.create_from_numpy_array("dst_field", src_arr)
        buffer = Field.create_generic("buffer", spatial_dimensions=1,
                                      field_type=FieldType.BUFFER, dtype=src_arr.dtype)

        pack_eqs = [Assignment(buffer.center(), src_field.center())]
        pack_types = {'src_field': gpu_src_arr.dtype, 'buffer': gpu_buffer_arr.dtype}

        config = CreateKernelConfig(target=pystencils.Target.GPU, data_type=pack_types)
        pack_ast = create_kernel(pack_eqs, config=config)

        pack_kernel = pack_ast.compile()
        pack_kernel(buffer=gpu_buffer_arr, src_field=gpu_src_arr)

        unpack_eqs = [Assignment(dst_field.center(), buffer.center())]
        unpack_types = {'dst_field': gpu_dst_arr.dtype, 'buffer': gpu_buffer_arr.dtype}

        config = CreateKernelConfig(target=pystencils.Target.GPU, data_type=unpack_types)
        unpack_ast = create_kernel(unpack_eqs, config=config)

        unpack_kernel = unpack_ast.compile()
        unpack_kernel(dst_field=gpu_dst_arr, buffer=gpu_buffer_arr)

        dst_arr = gpu_dst_arr.get()

        np.testing.assert_equal(src_arr, dst_arr)


def test_field_slice():
    """Tests (un)packing slices of a scalar field (from)to a buffer."""
    fields = _generate_fields()
    for d in ['N', 'S', 'NW', 'SW', 'TNW', 'B']:
        for (src_arr, gpu_src_arr, gpu_dst_arr, gpu_buffer_arr) in fields:
            # Extract slice from N direction of the field
            slice_dir = direction_string_to_offset(d, dim=len(src_arr.shape))
            pack_slice = get_slice_before_ghost_layer(slice_dir)
            unpack_slice = get_ghost_region_slice(slice_dir)

            src_field = Field.create_from_numpy_array("src_field", src_arr[pack_slice])
            dst_field = Field.create_from_numpy_array("dst_field", src_arr[unpack_slice])
            buffer = Field.create_generic("buffer", spatial_dimensions=1,
                                          field_type=FieldType.BUFFER, dtype=src_arr.dtype)

            pack_eqs = [Assignment(buffer.center(), src_field.center())]
            pack_types = {'src_field': gpu_src_arr.dtype, 'buffer': gpu_buffer_arr.dtype}

            config = CreateKernelConfig(target=pystencils.Target.GPU, data_type=pack_types)
            pack_ast = create_kernel(pack_eqs, config=config)

            pack_kernel = pack_ast.compile()
            pack_kernel(buffer=gpu_buffer_arr, src_field=gpu_src_arr[pack_slice])

            # Unpack into ghost layer of dst_field in N direction
            unpack_eqs = [Assignment(dst_field.center(), buffer.center())]
            unpack_types = {'dst_field': gpu_dst_arr.dtype, 'buffer': gpu_buffer_arr.dtype}

            config = CreateKernelConfig(target=pystencils.Target.GPU, data_type=unpack_types)
            unpack_ast = create_kernel(unpack_eqs, config=config)

            unpack_kernel = unpack_ast.compile()
            unpack_kernel(buffer=gpu_buffer_arr, dst_field=gpu_dst_arr[unpack_slice])

            dst_arr = gpu_dst_arr.get()

            np.testing.assert_equal(src_arr[pack_slice], dst_arr[unpack_slice])


def test_all_cell_values():
    """Tests (un)packing all cell values of the a field (from)to a buffer."""
    num_cell_values = 7
    fields = _generate_fields(stencil_directions=num_cell_values)
    for (src_arr, gpu_src_arr, gpu_dst_arr, gpu_buffer_arr) in fields:
        src_field = Field.create_from_numpy_array("src_field", gpu_src_arr, index_dimensions=1)
        dst_field = Field.create_from_numpy_array("dst_field", gpu_src_arr, index_dimensions=1)
        buffer = Field.create_generic("buffer", spatial_dimensions=1, index_dimensions=1,
                                      field_type=FieldType.BUFFER, dtype=gpu_src_arr.dtype)

        pack_eqs = []
        # Since we are packing all cell values for all cells, then
        # the buffer index is equivalent to the field index
        for idx in range(num_cell_values):
            eq = Assignment(buffer(idx), src_field(idx))
            pack_eqs.append(eq)

        pack_types = {'src_field': gpu_src_arr.dtype, 'buffer': gpu_buffer_arr.dtype}

        config = CreateKernelConfig(target=pystencils.Target.GPU, data_type=pack_types)
        pack_code = create_kernel(pack_eqs, config=config)
        pack_kernel = pack_code.compile()

        pack_kernel(buffer=gpu_buffer_arr, src_field=gpu_src_arr)

        unpack_eqs = []

        for idx in range(num_cell_values):
            eq = Assignment(dst_field(idx), buffer(idx))
            unpack_eqs.append(eq)

        unpack_types = {'dst_field': gpu_dst_arr.dtype, 'buffer': gpu_buffer_arr.dtype}

        config = CreateKernelConfig(target=pystencils.Target.GPU, data_type=unpack_types)
        unpack_ast = create_kernel(unpack_eqs, config=config)
        unpack_kernel = unpack_ast.compile()
        unpack_kernel(buffer=gpu_buffer_arr, dst_field=gpu_dst_arr)

        dst_arr = gpu_dst_arr.get()

        np.testing.assert_equal(src_arr, dst_arr)


def test_subset_cell_values():
    """Tests (un)packing a subset of cell values of the a field (from)to a buffer."""
    num_cell_values = 7
    # Cell indices of the field to be (un)packed (from)to the buffer
    cell_indices = [1, 3, 5, 6]
    fields = _generate_fields(stencil_directions=num_cell_values)
    for (src_arr, gpu_src_arr, gpu_dst_arr, gpu_buffer_arr) in fields:
        src_field = Field.create_from_numpy_array("src_field", gpu_src_arr, index_dimensions=1)
        dst_field = Field.create_from_numpy_array("dst_field", gpu_src_arr, index_dimensions=1)
        buffer = Field.create_generic("buffer", spatial_dimensions=1, index_dimensions=1,
                                      field_type=FieldType.BUFFER, dtype=gpu_src_arr.dtype)

        pack_eqs = []
        # Since we are packing all cell values for all cells, then
        # the buffer index is equivalent to the field index
        for buffer_idx, cell_idx in enumerate(cell_indices):
            eq = Assignment(buffer(buffer_idx), src_field(cell_idx))
            pack_eqs.append(eq)

        pack_types = {'src_field': gpu_src_arr.dtype, 'buffer': gpu_buffer_arr.dtype}
        config = CreateKernelConfig(target=pystencils.Target.GPU, data_type=pack_types)
        pack_ast = create_kernel(pack_eqs, config=config)
        pack_kernel = pack_ast.compile()
        pack_kernel(buffer=gpu_buffer_arr, src_field=gpu_src_arr)

        unpack_eqs = []

        for buffer_idx, cell_idx in enumerate(cell_indices):
            eq = Assignment(dst_field(cell_idx), buffer(buffer_idx))
            unpack_eqs.append(eq)

        unpack_types = {'dst_field': gpu_dst_arr.dtype, 'buffer': gpu_buffer_arr.dtype}
        config = CreateKernelConfig(target=pystencils.Target.GPU, data_type=unpack_types)
        unpack_ast = create_kernel(unpack_eqs, config=config)
        unpack_kernel = unpack_ast.compile()

        unpack_kernel(buffer=gpu_buffer_arr, dst_field=gpu_dst_arr)

        dst_arr = gpu_dst_arr.get()

        mask_arr = np.ma.masked_where((src_arr - dst_arr) != 0, src_arr)
        np.testing.assert_equal(dst_arr, mask_arr.filled(int(0)))


def test_field_layouts():
    num_cell_values = 7
    for layout_str in ['numpy', 'fzyx', 'zyxf', 'reverse_numpy']:
        fields = _generate_fields(stencil_directions=num_cell_values, layout=layout_str)
        for (src_arr, gpu_src_arr, gpu_dst_arr, gpu_buffer_arr) in fields:
            src_field = Field.create_from_numpy_array("src_field", gpu_src_arr, index_dimensions=1)
            dst_field = Field.create_from_numpy_array("dst_field", gpu_src_arr, index_dimensions=1)
            buffer = Field.create_generic("buffer", spatial_dimensions=1, index_dimensions=1,
                                          field_type=FieldType.BUFFER, dtype=src_arr.dtype)

            pack_eqs = []
            # Since we are packing all cell values for all cells, then
            # the buffer index is equivalent to the field index
            for idx in range(num_cell_values):
                eq = Assignment(buffer(idx), src_field(idx))
                pack_eqs.append(eq)

            pack_types = {'src_field': gpu_src_arr.dtype, 'buffer': gpu_buffer_arr.dtype}
            config = CreateKernelConfig(target=pystencils.Target.GPU, data_type=pack_types)
            pack_ast = create_kernel(pack_eqs, config=config)
            pack_kernel = pack_ast.compile()

            pack_kernel(buffer=gpu_buffer_arr, src_field=gpu_src_arr)

            unpack_eqs = []

            for idx in range(num_cell_values):
                eq = Assignment(dst_field(idx), buffer(idx))
                unpack_eqs.append(eq)

            unpack_types = {'dst_field': gpu_dst_arr.dtype, 'buffer': gpu_buffer_arr.dtype}
            config = CreateKernelConfig(target=pystencils.Target.GPU, data_type=unpack_types)
            unpack_ast = create_kernel(unpack_eqs, config=config)
            unpack_kernel = unpack_ast.compile()

            unpack_kernel(buffer=gpu_buffer_arr, dst_field=gpu_dst_arr)



from collections import namedtuple
from typing import Any, Callable, Optional, Sequence

import sympy as sp

from pystencils.simp.assignment_collection import AssignmentCollection


class SimplificationStrategy:
    """A simplification strategy is an ordered collection of simplification rules.

    Each simplification is a function taking an assignment collection, and returning a new simplified
    assignment collection. The strategy can nicely print intermediate simplification stages and results
    to Jupyter notebooks.
    """

    def __init__(self):
        self._rules = []

    def add(self, rule: Callable[[AssignmentCollection], AssignmentCollection]) -> None:
        """Adds the given simplification rule to the end of the collection.

        Args:
            rule: function that rewrites/simplifies an assignment collection
        """
        self._rules.append(rule)

    @property
    def rules(self):
        return self._rules

    def apply(self, assignment_collection: AssignmentCollection) -> AssignmentCollection:
        """Runs all rules on the given assignment collection."""
        for t in self._rules:
            assignment_collection = t(assignment_collection)
        return assignment_collection

    def __call__(self, assignment_collection: AssignmentCollection) -> AssignmentCollection:
        """Same as apply"""
        return self.apply(assignment_collection)

    def create_simplification_report(self, assignment_collection: AssignmentCollection) -> Any:
        """Creates a report to be displayed as HTML in a Jupyter notebook.

        The simplification report contains the number of operations at each simplification stage together
        with the run-time the simplification took.
        """

        ReportElement = namedtuple('ReportElement', ['simplificationName', 'runtime', 'adds', 'muls', 'divs', 'total'])

        class Report:
            def __init__(self):
                self.elements = []

            def add(self, element):
                self.elements.append(element)

            def __str__(self):
                try:
                    import tabulate
                    return tabulate(self.elements, headers=['Name', 'Runtime', 'Adds', 'Muls', 'Divs', 'Total'])
                except ImportError:
                    result = "Name, Adds, Muls, Divs, Runtime\n"
                    for e in self.elements:
                        result += ",".join([str(tuple_item) for tuple_item in e]) + "\n"
                    return result

            def _repr_html_(self):
                html_table = '<table style="border:none">'
                html_table += "<tr><th>Name</th>" \
                              "<th>Runtime</th>" \
                              "<th>Adds</th>" \
                              "<th>Muls</th>" \
                              "<th>Divs</th>" \
                              "<th>Total</th></tr>"
                line = "<tr><td>{simplificationName}</td>" \
                       "<td>{runtime}</td> <td>{adds}</td> <td>{muls}</td> <td>{divs}</td>  <td>{total}</td> </tr>"

                for e in self.elements:
                    # noinspection PyProtectedMember
                    html_table += line.format(**e._asdict())
                html_table += "</table>"
                return html_table

        import timeit
        report = Report()
        op = assignment_collection.operation_count
        total = op['adds'] + op['muls'] + op['divs']
        report.add(ReportElement("OriginalTerm", '-', op['adds'], op['muls'], op['divs'], total))
        for t in self._rules:
            start_time = timeit.default_timer()
            assignment_collection = t(assignment_collection)
            end_time = timeit.default_timer()
            op = assignment_collection.operation_count
            time_str = f"{(end_time - start_time) * 1000:.2f} ms"
            total = op['adds'] + op['muls'] + op['divs']
            report.add(ReportElement(t.__name__, time_str, op['adds'], op['muls'], op['divs'], total))
        return report

    def show_intermediate_results(self, assignment_collection: AssignmentCollection,
                                  symbols: Optional[Sequence[sp.Symbol]] = None) -> Any:
        """Shows the assignment collection after the application of each rule as HTML report for Jupyter notebook.

        Args:
            assignment_collection: the collection to apply the rules to
            symbols: if not None, only the assignments are shown that have one of these symbols as left hand side
        """
        class IntermediateResults:
            def __init__(self, strategy, collection, restrict_symbols):
                self.strategy = strategy
                self.assignment_collection = collection
                self.restrict_symbols = restrict_symbols

            def __str__(self):
                def print_assignment_collection(title, c):
                    text = title
                    if self.restrict_symbols:
                        text += "\n".join([str(e) for e in c.new_filtered(self.restrict_symbols).main_assignments])
                    else:
                        text += (" " * 3 + (" " * 3).join(str(c).splitlines(True)))
                    return text

                result = print_assignment_collection("Initial Version", self.assignment_collection)
                collection = self.assignment_collection
                for rule in self.strategy.rules:
                    collection = rule(collection)
                    result += print_assignment_collection(rule.__name__, collection)
                return result

            def _repr_html_(self):
                def print_assignment_collection(title, c):
                    text = f'<h5 style="padding-bottom:10px">{title}</h5> <div style="padding-left:20px;">'
                    if self.restrict_symbols:
                        text += "\n".join(["$$" + sp.latex(e) + '$$'
                                           for e in c.new_filtered(self.restrict_symbols).main_assignments])
                    else:
                        # noinspection PyProtectedMember
                        text += c._repr_html_()
                    text += "</div>"
                    return text

                result = print_assignment_collection("Initial Version", self.assignment_collection)
                collection = self.assignment_collection
                for rule in self.strategy.rules:
                    collection = rule(collection)
                    result += print_assignment_collection(rule.__name__, collection)
                return result

        return IntermediateResults(self, assignment_collection, symbols)

    def __repr__(self):
        result = "Simplification Strategy:\n"
        for t in self._rules:
            result += f" - {t.__name__}\n"
        return result



from pystencils.gpucuda.cudajit import make_python_function
from pystencils.gpucuda.kernelcreation import create_cuda_kernel, created_indexed_cuda_kernel

from .indexing import AbstractIndexing, BlockIndexing, LineIndexing

__all__ = ['create_cuda_kernel', 'created_indexed_cuda_kernel', 'make_python_function',
           'AbstractIndexing', 'BlockIndexing', 'LineIndexing']



import warnings
from copy import copy
from collections import defaultdict
from dataclasses import dataclass, field
from types import MappingProxyType
from typing import Union, Tuple, List, Dict, Callable, Any

from pystencils import Target, Backend, Field
from pystencils.typing.typed_sympy import BasicType

import numpy as np


# TODO: CreateKernelConfig is bloated think of more classes better usage, factory whatever ...
# Proposition: CreateKernelConfigs Classes for different targets?
@dataclass
class CreateKernelConfig:
    """
    **Below all parameters for the CreateKernelConfig are explained**
    """
    target: Target = Target.CPU
    """
    All targets are defined in :class:`pystencils.enums.Target`
    """
    backend: Backend = None
    """
    All backends are defined in :class:`pystencils.enums.Backend`
    """
    function_name: str = 'kernel'
    """
    Name of the generated function - only important if generated code is written out
    """
    # TODO Sane defaults: config should check that the datatype is a Numpy type
    # TODO Sane defaults: QoL default_number_float and default_number_int should be data_type if they are not specified
    data_type: Union[str, Dict[str, BasicType]] = 'float64'
    """
    Data type used for all untyped symbols (i.e. non-fields), can also be a dict from symbol name to type
    """
    default_number_float: Union[str, np.dtype, BasicType] = 'float64'
    """
    Data type used for all untyped floating point numbers (i.e. 0.5)
    """
    default_number_int: Union[str, np.dtype, BasicType] = 'int64'
    """
    Data type used for all untyped integer numbers (i.e. 1)
    """
    iteration_slice: Tuple = None
    """
    Rectangular subset to iterate over, if not specified the complete non-ghost layer part of the field is iterated over
    """
    ghost_layers: Union[bool, int, List[Tuple[int]]] = None
    """
    A single integer specifies the ghost layer count at all borders, can also be a sequence of
    pairs ``[(x_lower_gl, x_upper_gl), .... ]``. These layers are excluded from the iteration.
    If left to default, the number of ghost layers is determined automatically from the assignments.
    """
    cpu_openmp: Union[bool, int] = False
    """
    `True` or number of threads for OpenMP parallelization, `False` for no OpenMP. If set to `True`, the maximum number
    of available threads will be chosen.
    """
    cpu_vectorize_info: Dict = None
    """
    A dictionary with keys, 'vector_instruction_set', 'assume_aligned' and 'nontemporal'
    for documentation of these parameters see vectorize function. Example:
    '{'instruction_set': 'avx512', 'assume_aligned': True, 'nontemporal':True}'
    """
    cpu_blocking: Tuple[int] = None
    """
    A tuple of block sizes or `None` if no blocking should be applied
    """
    omp_single_loop: bool = True
    """
    If OpenMP is active: whether multiple outer loops are permitted
    """
    gpu_indexing: str = 'block'
    """
    Either 'block' or 'line' , or custom indexing class, see `pystencils.gpucuda.AbstractIndexing`
    """
    gpu_indexing_params: MappingProxyType = field(default=MappingProxyType({}))
    """
    Dict with indexing parameters (constructor parameters of indexing class)
    e.g. for 'block' one can specify '{'block_size': (20, 20, 10) }'.
    """
    # TODO Markus rework this docstring
    default_assignment_simplifications: bool = False
    """
    If `True` default simplifications are first performed on the Assignments. If problems occur during the
    simplification a warning will be thrown.
    Furthermore, it is essential to know that this is a two-stage process. The first stage of the process acts
    on the level of the `pystencils.AssignmentCollection`.  In this part,
    `pystencil.simp.create_simplification_strategy` from pystencils.simplificationfactory will be used to
    apply optimisations like insertion of constants to
    remove pressure from the registers. Thus the first part of the optimisations can only be executed if
    an `AssignmentCollection` is passed. The second part of the optimisation acts on the level of each Assignment
    individually. In this stage, all optimisations from `sympy.codegen.rewriting.optims_c99` are applied
    to each Assignment. Thus this stage can also be applied if a list of Assignments is passed.
    """
    cpu_prepend_optimizations: List[Callable] = field(default_factory=list)
    """
    List of extra optimizations to perform first on the AST.
    """
    use_auto_for_assignments: bool = False
    """
    If set to `True`, auto can be used in the generated code for data types. This makes the type system more robust.
    """
    index_fields: List[Field] = None
    """
    List of index fields, i.e. 1D fields with struct data type. If not `None`, `create_index_kernel`
    instead of `create_domain_kernel` is used.
    """
    coordinate_names: Tuple[str, Any] = ('x', 'y', 'z')
    """
    Name of the coordinate fields in the struct data type.
    """
    allow_double_writes: bool = False
    """
    If True, don't check if every field is only written at a single location. This is required
    for example for kernels that are compiled with loop step sizes > 1, that handle multiple 
    cells at once. Use with care!
    """
    skip_independence_check: bool = False
    """
    Don't check that loop iterations are independent. This is needed e.g. for 
    periodicity kernel, that access the field outside the iteration bounds. Use with care!
    """

    class DataTypeFactory:
        """Because of pickle, we need to have a nested class, instead of a lambda in __post_init__"""
        def __init__(self, dt):
            self.dt = dt

        def __call__(self):
            return BasicType(self.dt)

    def __post_init__(self):
        # ----  Legacy parameters
        # TODO Sane defaults: Check for abmigous types like "float", python float, which are dangerous for users
        if isinstance(self.target, str):
            new_target = Target[self.target.upper()]
            warnings.warn(f'Target "{self.target}" as str is deprecated. Use {new_target} instead',
                          category=DeprecationWarning)
            self.target = new_target
        # ---- Auto Backend
        if not self.backend:
            if self.target == Target.CPU:
                self.backend = Backend.C
            elif self.target == Target.GPU:
                self.backend = Backend.CUDA
            else:
                raise NotImplementedError(f'Target {self.target} has no default backend')

        #  Normalise data types
        if not isinstance(self.data_type, dict):
            dt = copy(self.data_type)  # The copy is necessary because BasicType has sympy shinanigans
            self.data_type = defaultdict(self.DataTypeFactory(dt))
        if not isinstance(self.default_number_float, BasicType):
            self.default_number_float = BasicType(self.default_number_float)
        if not isinstance(self.default_number_int, BasicType):
            self.default_number_int = BasicType(self.default_number_int)



import copy
import numpy as np
import sympy as sp

from pystencils.typing import TypedSymbol, CastFunc
from pystencils.astnodes import LoopOverCoordinate
from pystencils.backends.cbackend import CustomCodeNode
from pystencils.sympyextensions import fast_subs


class RNGBase(CustomCodeNode):

    id = 0

    def __init__(self, dim, time_step=TypedSymbol("time_step", np.uint32), offsets=None, keys=None):
        if keys is None:
            keys = (0,) * self._num_keys
        if offsets is None:
            offsets = (0,) * dim
        if len(keys) != self._num_keys:
            raise ValueError(f"Provided {len(keys)} keys but need {self._num_keys}")
        if len(offsets) != dim:
            raise ValueError(f"Provided {len(offsets)} offsets but need {dim}")
        coordinates = [LoopOverCoordinate.get_loop_counter_symbol(i) + offsets[i] for i in range(dim)]
        if dim < 3:
            coordinates.append(0)

        self._args = sp.sympify([time_step, *coordinates, *keys])
        self.result_symbols = tuple(TypedSymbol(f'random_{self.id}_{i}', self._data_type)
                                    for i in range(self._num_vars))
        symbols_read = set.union(*[s.atoms(sp.Symbol) for s in self.args])
        super().__init__("", symbols_read=symbols_read, symbols_defined=self.result_symbols)

        self.headers = [f'"{self._name.split("_")[0]}_rand.h"']

        RNGBase.id += 1

    @property
    def args(self):
        return self._args

    def fast_subs(self, subs_dict, skip):
        rng = copy.deepcopy(self)
        rng._args = [fast_subs(a, subs_dict, skip) for a in rng._args]
        return rng

    def get_code(self, dialect, vector_instruction_set, print_arg):
        code = "\n"
        for r in self.result_symbols:
            if vector_instruction_set and not self.args[1].atoms(CastFunc):
                # this vector RNG has become scalar through substitution
                code += f"{r.dtype} {r.name};\n"
            else:
                code += f"{vector_instruction_set[r.dtype.c_name] if vector_instruction_set else r.dtype} " + \
                        f"{r.name};\n"
        args = [print_arg(a) for a in self.args] + ['' + r.name for r in self.result_symbols]
        code += (self._name + "(" + ", ".join(args) + ");\n")
        return code

    def __repr__(self):
        return ", ".join([str(s) for s in self.result_symbols]) + " \\leftarrow " + \
            self._name.capitalize() + "_RNG(" + ", ".join([str(a) for a in self.args]) + ")"


class PhiloxTwoDoubles(RNGBase):
    _name = "philox_double2"
    _data_type = np.float64
    _num_vars = 2
    _num_keys = 2


class PhiloxFourFloats(RNGBase):
    _name = "philox_float4"
    _data_type = np.float32
    _num_vars = 4
    _num_keys = 2


class AESNITwoDoubles(RNGBase):
    _name = "aesni_double2"
    _data_type = np.float64
    _num_vars = 2
    _num_keys = 4


class AESNIFourFloats(RNGBase):
    _name = "aesni_float4"
    _data_type = np.float32
    _num_vars = 4
    _num_keys = 4


def random_symbol(assignment_list, dim, seed=TypedSymbol("seed", np.uint32), rng_node=PhiloxTwoDoubles,
                  time_step=TypedSymbol("time_step", np.uint32), offsets=None):
    """Return a symbol generator for random numbers
    
    Args:
        assignment_list: the subexpressions member of an AssignmentCollection, into which helper variables assignments
                         will be inserted
        dim: 2 or 3 for two or three spatial dimensions
        seed: an integer or TypedSymbol(..., np.uint32) to seed the random number generator. If you create multiple
              symbol generators, please pass them different seeds so you don't get the same stream of random numbers!
        rng_node: which random number generator to use (PhiloxTwoDoubles, PhiloxFourFloats, AESNITwoDoubles,
                  AESNIFourFloats).
        time_step: TypedSymbol(..., np.uint32) that indicates the number of the current time step
        offsets: tuple of offsets (constant integers or TypedSymbol(..., np.uint32)) that give the global coordinates
                 of the local origin
    """
    counter = 0
    while True:
        keys = (counter, seed) + (0,) * (rng_node._num_keys - 2)
        node = rng_node(dim, keys=keys, time_step=time_step, offsets=offsets)
        inserted = False
        for symbol in node.result_symbols:
            if not inserted:
                assignment_list.insert(0, node)
                inserted = True
            yield symbol
        counter += 1



from pystencils.cache import sharedmethodcache


class Fib:

    def __init__(self):
        self.fib_rec_called = 0
        self.fib_iter_called = 0

    @sharedmethodcache("fib_cache")
    def fib_rec(self, n):
        self.fib_rec_called += 1
        return 1 if n <= 1 else self.fib_rec(n-1) + self.fib_rec(n-2)

    @sharedmethodcache("fib_cache")
    def fib_iter(self, n):
        self.fib_iter_called += 1
        f1, f2 = 0, 1
        for i in range(n):
            f2 = f1 + f2
            f1 = f2 - f1
        return f2


def test_fib_memoization_1():
    fib = Fib()

    assert "fib_cache" not in fib.__dict__

    f13 = fib.fib_rec(13)
    assert fib.fib_rec_called == 14
    assert "fib_cache" in fib.__dict__
    assert fib.fib_cache[(13,)] == f13

    for k in range(14):
        #   fib_iter should use cached results from fib_rec
        fib.fib_iter(k)
    
    assert fib.fib_iter_called == 0


def test_fib_memoization_2():
    fib = Fib()

    f11 = fib.fib_iter(11)
    f12 = fib.fib_iter(12)

    assert fib.fib_iter_called == 2

    f13 = fib.fib_rec(13)

    #   recursive calls should be cached
    assert fib.fib_rec_called == 1


class Triad:
    
    def __init__(self):
        self.triad_called = 0

    @sharedmethodcache("triad_cache")
    def triad(self, a, b, c=0):
        """Computes the triad a*b+c."""
        self.triad_called += 1
        return a * b + c


def test_triad_memoization():
    triad = Triad()

    assert triad.triad.__doc__ == "Computes the triad a*b+c."

    t = triad.triad(12, 4, 15)
    assert triad.triad_called == 1
    assert triad.triad_cache[(12, 4, 15)] == t

    t = triad.triad(12, 4, c=15)
    assert triad.triad_called == 2
    assert triad.triad_cache[(12, 4, 'c', 15)] == t

    t = triad.triad(12, 4, 15)
    assert triad.triad_called == 2

    t = triad.triad(12, 4, c=15)
    assert triad.triad_called == 2



import numpy as np
import sympy as sp

import pystencils as ps
from pystencils.jupyter import make_imshow_animation, display_animation, set_display_mode
import pystencils.plot as plt

__all__ = ['sp', 'np', 'ps', 'plt', 'make_imshow_animation', 'display_animation', 'set_display_mode']



import base64
from tempfile import NamedTemporaryFile

import matplotlib.animation as animation
import sympy as sp
from IPython.display import HTML

import pystencils.plot as plt

VIDEO_TAG = """<video controls width="80%">
 <source src="data:video/x-m4v;base64,{0}" type="video/mp4">
 Your browser does not support the video tag.
</video>"""


def __animation_to_html(animation, fps):
    if not hasattr(animation, 'encoded_video'):
        with NamedTemporaryFile(suffix='.mp4') as f:
            animation.save(f.name, fps=fps, extra_args=['-vcodec', 'libx264', '-pix_fmt',
                                                        'yuv420p', '-profile:v', 'baseline', '-level', '3.0'])
            video = open(f.name, "rb").read()
        animation.encoded_video = base64.b64encode(video).decode('ascii')

    return VIDEO_TAG.format(animation.encoded_video)


def make_imshow_animation(grid, grid_update_function, frames=90, **_):
    from functools import partial
    fig = plt.figure()
    im = plt.imshow(grid, interpolation='none')

    def update_figure(*_, **kwargs):
        image = kwargs['image']
        image = grid_update_function(image)
        im.set_array(image)
        return im,

    return animation.FuncAnimation(fig, partial(update_figure, image=grid), frames=frames)


# -------   Version 1: Embed the animation as HTML5 video --------- ----------------------------------

def display_as_html_video(animation, fps=30, show=True, **_):
    try:
        plt.close(animation._fig)
        res = __animation_to_html(animation, fps)
        if show:
            return HTML(res)
        else:
            return HTML("")
    except KeyboardInterrupt:
        pass


# -------   Version 2: Animation is shown in extra matplotlib window ----------------------------------


def display_in_extra_window(*_, **__):
    fig = plt.gcf()
    try:
        fig.canvas.manager.window.raise_()
    except Exception:
        pass
    plt.show()


# -------   Version 3: Animation is shown in images that are updated directly in website --------------

def display_as_html_image(animation, show=True, *args, **kwargs):
    from IPython import display

    try:
        if show:
            animation._init_draw()
        for _ in animation.frame_seq:
            if show:
                fig = plt.gcf()
                display.display(fig)
            animation._step()
            if show:
                display.clear_output(wait=True)
    except KeyboardInterrupt:
        display.clear_output(wait=False)


# Dispatcher

animation_display_mode = 'image_update'
display_animation_func = None


def display_animation(*args, **kwargs):
    from IPython import get_ipython
    ipython = get_ipython()
    if not ipython:
        return

    if not display_animation_func:
        raise Exception("Call set_display_mode first")
    return display_animation_func(*args, **kwargs)


def set_display_mode(mode):
    from IPython import get_ipython
    ipython = get_ipython()
    if not ipython:
        return
    global animation_display_mode
    global display_animation_func
    animation_display_mode = mode
    if animation_display_mode == 'video':
        ipython.magic("matplotlib inline")
        display_animation_func = display_as_html_video
    elif animation_display_mode == 'window':
        ipython.magic("matplotlib qt")
        display_animation_func = display_in_extra_window
    elif animation_display_mode == 'image_update':
        ipython.magic("matplotlib inline")
        display_animation_func = display_as_html_image
    else:
        raise Exception("Unknown mode. Available modes 'image_update', 'video' and 'window' ")


def activate_ipython():
    from IPython import get_ipython
    ipython = get_ipython()
    if ipython:
        set_display_mode('image_update')
        ipython.magic("config InlineBackend.rc = { }")
        ipython.magic("matplotlib inline")
        plt.rc('figure', figsize=(16, 6))
        sp.init_printing()


activate_ipython()



"""
This module extends the pyplot module with functions to show scalar and vector fields in the usual
simulation coordinate system (y-axis goes up), instead of the "image coordinate system" (y axis goes down) that
matplotlib normally uses.
"""
import warnings
from itertools import cycle

from matplotlib.pyplot import *


def vector_field(array, step=2, **kwargs):
    """Plots given vector field as quiver (arrow) plot.

    Args:
        array: numpy array with 3 dimensions, first two are spatial x,y coordinate, the last
               coordinate should have shape 2 and stores the 2 velocity components
        step: plots only every steps's cell, increase the step for high resolution arrays
        kwargs: keyword arguments passed to :func:`matplotlib.pyplot.quiver`

    Returns:
        quiver plot object
    """
    assert len(array.shape) == 3, "Wrong shape of array - did you forget to slice your 3D domain first?"
    assert array.shape[2] == 2, "Last array dimension is expected to store 2D vectors"
    vel_n = array.swapaxes(0, 1)
    res = quiver(vel_n[::step, ::step, 0], vel_n[::step, ::step, 1], **kwargs)
    axis('equal')
    return res


def vector_field_magnitude(array, **kwargs):
    """Plots the magnitude of a vector field as colormap.

    Args:
        array: numpy array with 3 dimensions, first two are spatial x,y coordinate, the last
               coordinate should have shape 2 and stores the 2 velocity components
        kwargs: keyword arguments passed to :func:`matplotlib.pyplot.imshow`

    Returns:
        imshow object
    """
    assert len(array.shape) == 3, "Wrong shape of array - did you forget to slice your 3D domain first?"
    assert array.shape[2] in (2, 3), "Wrong size of the last coordinate. Has to be a 2D or 3D vector field."
    from numpy.linalg import norm
    norm = norm(array, axis=2, ord=2)
    if hasattr(array, 'mask'):
        norm = np.ma.masked_array(norm, mask=array.mask[:, :, 0])
    return scalar_field(norm, **kwargs)


def scalar_field(array, **kwargs):
    """Plots field values as colormap.

    Works just as imshow, but uses coordinate system where second coordinate (y) points upwards.

    Args:
        array: two dimensional numpy array
        kwargs: keyword arguments passed to :func:`matplotlib.pyplot.imshow`

    Returns:
        imshow object
    """
    import numpy
    array = numpy.swapaxes(array, 0, 1)
    res = imshow(array, origin='lower', **kwargs)
    axis('equal')
    return res


def scalar_field_surface(array, **kwargs):
    """Plots scalar field as 3D surface

    Args:
        array: the two dimensional numpy array to plot
        kwargs: keyword arguments passed to :func:`mpl_toolkits.mplot3d.Axes3D.plot_surface`
    """
    from mpl_toolkits.mplot3d import Axes3D
    from matplotlib import cm

    fig = gcf()
    ax = fig.add_subplot(111, projection='3d')
    x, y = np.meshgrid(np.arange(array.shape[0]), np.arange(array.shape[1]), indexing='ij')
    kwargs.setdefault('rstride', 2)
    kwargs.setdefault('cstride', 2)
    kwargs.setdefault('color', 'b')
    kwargs.setdefault('cmap', cm.coolwarm)
    return ax.plot_surface(x, y, array, **kwargs)


def scalar_field_alpha_value(array, color, clip=False, **kwargs):
    """Plots an image with same color everywhere, using the array values as transparency.

    Array is supposed to have values between 0 and 1 (if this is not the case it is normalized).
    An image is plotted that has the same color everywhere, the passed array determines the transparency.
    Regions where the array is 1  are fully opaque, areas with 0 are fully transparent.

    Args:
        array: 2D array with alpha values
        color: fill color
        clip: if True, all values in the array larger than 1 are set to 1, all values smaller than 0 are set to zero
              if False, the array is linearly scaled to the [0, 1] interval
        **kwargs: arguments passed to imshow

    Returns:
        imshow object
    """
    import numpy
    import matplotlib
    assert len(array.shape) == 2, "Wrong shape of array - did you forget to slice your 3D domain first?"
    array = numpy.swapaxes(array, 0, 1)

    if clip:
        normalized_field = array.copy()
        normalized_field[normalized_field < 0] = 0
        normalized_field[normalized_field > 1] = 1
    else:
        minimum, maximum = numpy.min(array), numpy.max(array)
        normalized_field = (array - minimum) / (maximum - minimum)

    color = matplotlib.colors.to_rgba(color)
    field_to_plot = numpy.empty(array.shape + (4,))
    # set the complete array to the color
    for i in range(3):
        field_to_plot[:, :, i] = color[i]
    # only the alpha channel varies using the array values
    field_to_plot[:, :, 3] = normalized_field

    res = imshow(field_to_plot, origin='lower', **kwargs)
    axis('equal')
    return res


def scalar_field_contour(array, **kwargs):
    """Small wrapper around contour to transform the coordinate system.

    For details see  :func:`matplotlib.pyplot.imshow`
    """
    array = np.swapaxes(array, 0, 1)
    res = contour(array, **kwargs)
    axis('equal')
    return res


def multiple_scalar_fields(array, **kwargs):
    """Plots a 3D array by slicing the last dimension and creates on plot for each entry of the last dimension.

    Args:
        array: 3D array to plot.
        **kwargs: passed along to imshow
    """
    assert len(array.shape) == 3
    sub_plots = array.shape[-1]
    for i in range(sub_plots):
        subplot(1, sub_plots, i + 1)
        title(str(i))
        scalar_field(array[..., i], **kwargs)
        colorbar()


def phase_plot(phase_field: np.ndarray, linewidth=1.0, clip=True) -> None:
    """Plots a phase field array using the phase variables as alpha channel.

    Args:
        phase_field: array with len(shape) == 3, first two dimensions are spatial, the last one indexes the phase
                     components.
        linewidth: line width of the 0.5 contour lines that are drawn over the alpha blended phase images
        clip: see scalar_field_alpha_value function
    """
    color_cycle = cycle(['#fe0002', '#00fe00', '#0000ff', '#ffa800', '#f600ff'])

    assert len(phase_field.shape) == 3

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        for i in range(phase_field.shape[-1]):
            scalar_field_alpha_value(phase_field[..., i], next(color_cycle), clip=clip, interpolation='bilinear')
        if linewidth:
            for i in range(phase_field.shape[-1]):
                scalar_field_contour(phase_field[..., i], levels=[0.5], colors='k', linewidths=[linewidth])


def sympy_function(expr, x_values=None, **kwargs):
    """Plots the graph of a sympy term that depends on one symbol only.

    Args:
        expr: sympy term that depends on one symbol only, which is plotted on the x axis
        x_values: describes sampling of x axis. Possible values are:
                     * tuple of (start, stop) or (start, stop, nr_of_steps)
                     * None, then start=0, stop=1, nr_of_steps=100
                     * 1D numpy array with x values
        **kwargs: passed on to :func:`matplotlib.pyplot.plot`

    Returns:
        plot object
    """
    import sympy as sp
    if x_values is None:
        x_arr = np.linspace(0, 1, 100)
    elif type(x_values) is tuple:
        x_arr = np.linspace(*x_values)
    elif isinstance(x_values, np.ndarray):
        assert len(x_values.shape) == 1
        x_arr = x_values
    else:
        raise ValueError("Invalid value for parameter x_values")
    symbols = expr.atoms(sp.Symbol)
    assert len(symbols) == 1, "Sympy expression may only depend on one variable only. Depends on " + str(symbols)
    y_arr = sp.lambdify(symbols.pop(), expr)(x_arr)
    return plot(x_arr, y_arr, **kwargs)


# ------------------------------------------- Animations ---------------------------------------------------------------

def __scale_array(arr):
    from numpy.linalg import norm
    norm_arr = norm(arr, axis=2, ord=2)
    if isinstance(arr, np.ma.MaskedArray):
        norm_arr = np.ma.masked_array(norm_arr, arr.mask[..., 0])
    return arr / norm_arr.max()


def vector_field_animation(run_function, step=2, rescale=True, plot_setup_function=lambda *_: None,
                           plot_update_function=lambda *_: None, interval=200, frames=180, **kwargs):
    """Creates a matplotlib animation of a vector field using a quiver plot.

    Args:
        run_function: callable without arguments, returning a 2D vector field i.e. numpy array with len(shape)==3
        step: see documentation of vector_field function
        rescale: if True, the length of the arrows is rescaled in every time step
        plot_setup_function: optional callable with the quiver object as argument,
                             that can be used to set up the plot (title, legend,..)
        plot_update_function: optional callable with the quiver object as argument
                              that is called of the quiver object was updated
        interval: delay between frames in milliseconds (see matplotlib.FuncAnimation)
        frames: how many frames should be generated, see matplotlib.FuncAnimation
        **kwargs: passed to quiver plot

    Returns:
        matplotlib animation object
    """
    import matplotlib.animation as animation

    fig = gcf()
    im = None
    field = run_function()
    if rescale:
        field = __scale_array(field)
        kwargs.setdefault('scale', 0.6)
        kwargs.setdefault('angles', 'xy')
        kwargs.setdefault('scale_units', 'xy')

    quiver_plot = vector_field(field, step=step, **kwargs)
    plot_setup_function(quiver_plot)

    def update_figure(*_):
        f = run_function()
        f = np.swapaxes(f, 0, 1)
        if rescale:
            f = __scale_array(f)
        u, v = f[::step, ::step, 0], f[::step, ::step, 1]
        quiver_plot.set_UVC(u, v)
        plot_update_function(quiver_plot)
        return im,

    return animation.FuncAnimation(fig, update_figure, interval=interval, frames=frames)


def vector_field_magnitude_animation(run_function, plot_setup_function=lambda *_: None, rescale=False,
                                     plot_update_function=lambda *_: None, interval=30, frames=180, **kwargs):
    """Animation of a vector field, showing the magnitude as colormap.

    For arguments, see vector_field_animation
    """
    import matplotlib.animation as animation
    from numpy.linalg import norm

    fig = gcf()
    im = None
    field = run_function()
    if rescale:
        field = __scale_array(field)
    im = vector_field_magnitude(field, **kwargs)
    plot_setup_function(im)

    def update_figure(*_):
        f = run_function()
        if rescale:
            f = __scale_array(f)
        normed = norm(f, axis=2, ord=2)
        if hasattr(f, 'mask'):
            normed = np.ma.masked_array(normed, mask=f.mask[:, :, 0])
        normed = np.swapaxes(normed, 0, 1)
        im.set_array(normed)
        plot_update_function(im)
        return im,

    return animation.FuncAnimation(fig, update_figure, interval=interval, frames=frames)


def scalar_field_animation(run_function, plot_setup_function=lambda *_: None, rescale=True,
                           plot_update_function=lambda *_: None, interval=30, frames=180, **kwargs):
    """Animation of scalar field as colored image, see `scalar_field`."""
    import matplotlib.animation as animation

    fig = gcf()
    im = None
    field = run_function()
    if rescale:
        f_min, f_max = np.min(field), np.max(field)
        field = (field - f_min) / (f_max - f_min)
        im = scalar_field(field, vmin=0.0, vmax=1.0, **kwargs)
    else:
        im = scalar_field(field, **kwargs)
    plot_setup_function(im)

    def update_figure(*_):
        f = run_function()
        if rescale:
            f_min, f_max = np.min(f), np.max(f)
            f = (f - f_min) / (f_max - f_min)
        if hasattr(f, 'mask'):
            f = np.ma.masked_array(f, mask=f.mask[:, :])
        f = np.swapaxes(f, 0, 1)
        im.set_array(f)
        plot_update_function(im)
        return im,

    return animation.FuncAnimation(fig, update_figure, interval=interval, frames=frames)


def surface_plot_animation(run_function, frames=90, interval=30, zlim=None,  **kwargs):
    """Animation of scalar field as 3D plot."""
    from mpl_toolkits.mplot3d import Axes3D
    import matplotlib.animation as animation
    from matplotlib import cm
    fig = gcf()
    ax = fig.add_subplot(111, projection='3d')
    data = run_function()
    x, y = np.meshgrid(np.arange(data.shape[0]), np.arange(data.shape[1]), indexing='ij')
    kwargs.setdefault('rstride', 2)
    kwargs.setdefault('cstride', 2)
    kwargs.setdefault('color', 'b')
    kwargs.setdefault('cmap', cm.coolwarm)
    ax.plot_surface(x, y, data, **kwargs)
    if zlim is not None:
        ax.set_zlim(*zlim)

    def update_figure(*_):
        d = run_function()
        ax.clear()
        plot = ax.plot_surface(x, y, d, **kwargs)
        if zlim is not None:
            ax.set_zlim(*zlim)
        return plot,

    return animation.FuncAnimation(fig, update_figure, interval=interval, frames=frames, blit=False)



from typing import Union

import sympy as sp

import pystencils.astnodes as ast
from pystencils.simp.assignment_collection import AssignmentCollection
from pystencils.config import CreateKernelConfig
from pystencils.enums import Target, Backend
from pystencils.astnodes import Block, KernelFunction, LoopOverCoordinate, SympyAssignment
from pystencils.cpu.cpujit import make_python_function
from pystencils.typing import StructType, TypedSymbol, create_type
from pystencils.typing.transformations import add_types
from pystencils.field import Field, FieldType
from pystencils.node_collection import NodeCollection
from pystencils.transformations import (
    filtered_tree_iteration, get_base_buffer_index, get_optimal_loop_ordering, make_loop_over_domain,
    move_constants_before_loop, parse_base_pointer_info, resolve_buffer_accesses,
    resolve_field_accesses, split_inner_loop)


def create_kernel(assignments: Union[AssignmentCollection, NodeCollection],
                  config: CreateKernelConfig) -> KernelFunction:
    """Creates an abstract syntax tree for a kernel function, by taking a list of update rules.

    Loops are created according to the field accesses in the equations.

    Args:
        assignments: list of sympy equations, containing accesses to :class:`pystencils.field.Field`.
        Defining the update rules of the kernel
        config: create kernel config

    Returns:
        AST node representing a function, that can be printed as C or CUDA code
    """
    function_name = config.function_name
    iteration_slice = config.iteration_slice
    ghost_layers = config.ghost_layers
    fields_written = assignments.bound_fields
    fields_read = assignments.rhs_fields

    split_groups = ()
    if 'split_groups' in assignments.simplification_hints:
        split_groups = assignments.simplification_hints['split_groups']
    assignments = assignments.all_assignments

    # TODO Cleanup: move add_types to create_domain_kernel or create_kernel
    assignments = add_types(assignments, config)

    all_fields = fields_read.union(fields_written)
    read_only_fields = set([f.name for f in fields_read - fields_written])

    buffers = set([f for f in all_fields if FieldType.is_buffer(f)])
    fields_without_buffers = all_fields - buffers

    body = ast.Block(assignments)
    loop_order = get_optimal_loop_ordering(fields_without_buffers)
    loop_node, ghost_layer_info = make_loop_over_domain(body, iteration_slice=iteration_slice,
                                                        ghost_layers=ghost_layers, loop_order=loop_order)
    ast_node = KernelFunction(loop_node, Target.CPU, Backend.C, compile_function=make_python_function,
                              ghost_layers=ghost_layer_info, function_name=function_name, assignments=assignments)

    if split_groups:
        type_info = config.data_type

        def type_symbol(term):
            if isinstance(term, Field.Access) or isinstance(term, TypedSymbol):
                return term
            elif isinstance(term, sp.Symbol):
                if isinstance(type_info, str) or not hasattr(type_info, '__getitem__'):
                    return TypedSymbol(term.name, create_type(type_info))
                else:
                    return TypedSymbol(term.name, type_info[term.name])
            else:
                raise ValueError("Term has to be field access or symbol")

        typed_split_groups = [[type_symbol(s) for s in split_group] for split_group in split_groups]
        split_inner_loop(ast_node, typed_split_groups)

    base_pointer_spec = [['spatialInner0'], ['spatialInner1']] if len(loop_order) >= 2 else [['spatialInner0']]
    base_pointer_info = {field.name: parse_base_pointer_info(base_pointer_spec, loop_order,
                                                             field.spatial_dimensions, field.index_dimensions)
                         for field in fields_without_buffers}

    buffer_base_pointer_info = {field.name: parse_base_pointer_info([['spatialInner0']], [0],
                                                                    field.spatial_dimensions, field.index_dimensions)
                                for field in buffers}
    base_pointer_info.update(buffer_base_pointer_info)

    if any(FieldType.is_buffer(f) for f in all_fields):
        resolve_buffer_accesses(ast_node, get_base_buffer_index(ast_node), read_only_fields)
    # TODO think about typing
    resolve_field_accesses(ast_node, read_only_fields, field_to_base_pointer_info=base_pointer_info)
    move_constants_before_loop(ast_node)
    return ast_node


def create_indexed_kernel(assignments: Union[AssignmentCollection, NodeCollection],
                          config: CreateKernelConfig) -> KernelFunction:
    """
    Similar to :func:`create_kernel`, but here not all cells of a field are updated but only cells with
    coordinates which are stored in an index field. This traversal method can e.g. be used for boundary handling.

    The coordinates are stored in a separate index_field, which is a one dimensional array with struct data type.
    This struct has to contain fields named 'x', 'y' and for 3D fields ('z'). These names are configurable with the
    'coordinate_names' parameter. The struct can have also other fields that can be read and written in the kernel, for
    example boundary parameters.

    Args:
        assignments: list of assignments
        config: Kernel configuration
    """
    function_name = config.function_name
    index_fields = config.index_fields
    coordinate_names = config.coordinate_names
    fields_written = assignments.bound_fields
    fields_read = assignments.rhs_fields

    assignments = assignments.all_assignments
    assignments = add_types(assignments, config)

    all_fields = fields_read.union(fields_written)

    for index_field in index_fields:
        index_field.field_type = FieldType.INDEXED
        assert FieldType.is_indexed(index_field)
        assert index_field.spatial_dimensions == 1, "Index fields have to be 1D"

    non_index_fields = [f for f in all_fields if f not in index_fields]
    spatial_coordinates = {f.spatial_dimensions for f in non_index_fields}
    assert len(spatial_coordinates) == 1, "Non-index fields do not have the same number of spatial coordinates"
    spatial_coordinates = list(spatial_coordinates)[0]

    def get_coordinate_symbol_assignment(name):
        for idx_field in index_fields:
            assert isinstance(idx_field.dtype, StructType), "Index fields have to have a struct data type"
            data_type = idx_field.dtype
            if data_type.has_element(name):
                rhs = idx_field[0](name)
                lhs = TypedSymbol(name, data_type.get_element_type(name))
                return SympyAssignment(lhs, rhs)
        raise ValueError(f"Index {name} not found in any of the passed index fields")

    coordinate_symbol_assignments = [get_coordinate_symbol_assignment(n)
                                     for n in coordinate_names[:spatial_coordinates]]
    coordinate_typed_symbols = [eq.lhs for eq in coordinate_symbol_assignments]
    assignments = coordinate_symbol_assignments + assignments

    # make 1D loop over index fields
    loop_body = Block([])
    loop_node = LoopOverCoordinate(loop_body, coordinate_to_loop_over=0, start=0, stop=index_fields[0].shape[0])

    for assignment in assignments:
        loop_body.append(assignment)

    function_body = Block([loop_node])
    ast_node = KernelFunction(function_body, Target.CPU, Backend.C, make_python_function,
                              ghost_layers=None, function_name=function_name, assignments=assignments)

    fixed_coordinate_mapping = {f.name: coordinate_typed_symbols for f in non_index_fields}

    read_only_fields = set([f.name for f in fields_read - fields_written])
    resolve_field_accesses(ast_node, read_only_fields, field_to_fixed_coordinates=fixed_coordinate_mapping)
    move_constants_before_loop(ast_node)
    return ast_node


def add_openmp(ast_node, schedule="static", num_threads=True, collapse=None, assume_single_outer_loop=True):
    """Parallelize the outer loop with OpenMP.

    Args:
        ast_node: abstract syntax tree created e.g. by :func:`create_kernel`
        schedule: OpenMP scheduling policy e.g. 'static' or 'dynamic'
        num_threads: explicitly specify number of threads
        collapse: number of nested loops to include in parallel region (see OpenMP collapse)
        assume_single_outer_loop: if True an exception is raised if multiple outer loops are detected for all but
                                  optimized staggered kernels the single-outer-loop assumption should be true
    """
    if not num_threads:
        return

    assert type(ast_node) is ast.KernelFunction
    body = ast_node.body
    threads_clause = "" if num_threads and isinstance(num_threads, bool) else f" num_threads({num_threads})"
    wrapper_block = ast.PragmaBlock('#pragma omp parallel' + threads_clause, body.take_child_nodes())
    body.append(wrapper_block)

    outer_loops = [l for l in filtered_tree_iteration(body, LoopOverCoordinate, stop_type=SympyAssignment)
                   if l.is_outermost_loop]
    assert outer_loops, "No outer loop found"
    if assume_single_outer_loop and len(outer_loops) > 1:
        raise ValueError("More than one outer loop found, only one outer loop expected")

    for loop_to_parallelize in outer_loops:
        try:
            loop_range = int(loop_to_parallelize.stop - loop_to_parallelize.start)
        except TypeError:
            loop_range = None

        if loop_range is not None and loop_range < num_threads and not collapse:
            contained_loops = [l for l in loop_to_parallelize.body.args if isinstance(l, LoopOverCoordinate)]
            if len(contained_loops) == 1:
                contained_loop = contained_loops[0]
                try:
                    contained_loop_range = int(contained_loop.stop - contained_loop.start)
                    if contained_loop_range > loop_range:
                        loop_to_parallelize = contained_loop
                except TypeError:
                    pass

        prefix = f"#pragma omp for schedule({schedule})"
        if collapse:
            prefix += f" collapse({collapse})"
        loop_to_parallelize.prefix_lines.append(prefix)



import numpy as np
import sympy as sp

from pystencils import Assignment, Field, TypedSymbol, create_kernel, make_slice
from pystencils.simp import sympy_cse_on_assignment_list


def test_sliced_iteration():
    size = (4, 4)
    src_arr = np.ones(size)
    dst_arr = np.zeros_like(src_arr)
    src_field = Field.create_from_numpy_array('src', src_arr)
    dst_field = Field.create_from_numpy_array('dst', dst_arr)

    a, b = sp.symbols("a b")
    update_rule = Assignment(dst_field[0, 0],
                             (a * src_field[0, 1] + a * src_field[0, -1] +
                              b * src_field[1, 0] + b * src_field[-1, 0]) / 4)

    x_end = TypedSymbol("x_end", "int")
    s = make_slice[1:x_end, 1]
    x_end_value = size[1] - 1
    kernel = create_kernel(sympy_cse_on_assignment_list([update_rule]), iteration_slice=s).compile()

    kernel(src=src_arr, dst=dst_arr, a=1.0, b=1.0, x_end=x_end_value)

    expected_result = np.zeros(size)
    expected_result[1:x_end_value, 1] = 1
    np.testing.assert_almost_equal(expected_result, dst_arr)



"""Module to generate stencil kernels in C or CUDA using sympy expressions and call them as Python functions"""
from .enums import Backend, Target
from . import fd
from . import stencil as stencil
from .assignment import Assignment, assignment_from_stencil
from pystencils.typing.typed_sympy import TypedSymbol
from .datahandling import create_data_handling
from .display_utils import get_code_obj, get_code_str, show_code, to_dot
from .field import Field, FieldType, fields
from .config import CreateKernelConfig
from .kernel_decorator import kernel, kernel_config
from .kernelcreation import create_kernel, create_staggered_kernel
from .simp import AssignmentCollection
from .slicing import make_slice
from .spatial_coordinates import x_, x_staggered, x_staggered_vector, x_vector, y_, y_staggered, z_, z_staggered
from .sympyextensions import SymbolCreator

__all__ = ['Field', 'FieldType', 'fields',
           'TypedSymbol',
           'make_slice',
           'CreateKernelConfig',
           'create_kernel', 'create_staggered_kernel',
           'Target', 'Backend',
           'show_code', 'to_dot', 'get_code_obj', 'get_code_str',
           'AssignmentCollection',
           'Assignment',
           'assignment_from_stencil',
           'SymbolCreator',
           'create_data_handling',
           'kernel', 'kernel_config',
           'x_', 'y_', 'z_',
           'x_staggered', 'y_staggered', 'z_staggered',
           'x_vector', 'x_staggered_vector',
           'fd',
           'stencil']

from ._version import get_versions

__version__ = get_versions()['version']
del get_versions



import pystencils


class KernelWrapper:
    """
    Light-weight wrapper around a compiled kernel.

    Can be called while still providing access to underlying AST.
    """

    def __init__(self, kernel, parameters, ast_node: pystencils.astnodes.KernelFunction):
        self.kernel = kernel
        self.parameters = parameters
        self.ast = ast_node
        self.num_regs = None

    def __call__(self, **kwargs):
        return self.kernel(**kwargs)

    @property
    def code(self):
        return pystencils.get_code_str(self.ast)



import sympy as sp
import pystencils as ps
import numpy as np
import pytest
from itertools import product
from pystencils.rng import random_symbol
from pystencils.astnodes import SympyAssignment
from pystencils.node_collection import NodeCollection


def advection_diffusion(dim: int):
    # parameters
    if dim == 2:
        L = (32, 32)
    elif dim == 3:
        L = (16, 16, 16)

    dh = ps.create_data_handling(domain_size=L, periodicity=True, default_target=ps.Target.CPU)

    n_field = dh.add_array('n', values_per_cell=1)
    j_field = dh.add_array('j', values_per_cell=3 ** dim // 2, field_type=ps.FieldType.STAGGERED_FLUX)
    velocity_field = dh.add_array('v', values_per_cell=dim)

    D = 0.0666
    time = 100

    def grad(f):
        return sp.Matrix([ps.fd.diff(f, i) for i in range(dim)])

    flux_eq = - D * grad(n_field)
    fvm_eq = ps.fd.FVM1stOrder(n_field, flux=flux_eq)

    vof_adv = ps.fd.VOF(j_field, velocity_field, n_field)

    # merge calculation of advection and diffusion terms
    flux = []
    for adv, div in zip(vof_adv, fvm_eq.discrete_flux(j_field)):
        assert adv.lhs == div.lhs
        flux.append(ps.Assignment(adv.lhs, adv.rhs + div.rhs))

    flux_kernel = ps.create_staggered_kernel(flux).compile()

    pde_kernel = ps.create_kernel(fvm_eq.discrete_continuity(j_field)).compile()

    sync_conc = dh.synchronization_function([n_field.name])

    # analytical density calculation
    def density(pos: np.ndarray, time: int, D: float):
        return (4 * np.pi * D * time)**(-dim / 2) * \
            np.exp(-np.sum(np.square(pos), axis=-1) / (4 * D * time))

    pos = np.zeros((*L, dim))
    xpos = np.arange(-L[0] // 2, L[0] // 2)
    ypos = np.arange(-L[1] // 2, L[1] // 2)

    if dim == 2:
        pos[..., 1], pos[..., 0] = np.meshgrid(xpos, ypos)
    elif dim == 3:
        zpos = np.arange(-L[2] // 2, L[2] // 2)
        pos[..., 2], pos[..., 1], pos[..., 0] = np.meshgrid(xpos, ypos, zpos)
    pos += 0.5

    def run(velocity: np.ndarray, time: int):
        dh.fill(n_field.name, np.nan, ghost_layers=True, inner_ghost_layers=True)
        dh.fill(j_field.name, np.nan, ghost_layers=True, inner_ghost_layers=True)

        # set initial values for velocity and density
        for i in range(dim):
            dh.fill(velocity_field.name, velocity[i], i, ghost_layers=True, inner_ghost_layers=True)
        dh.fill(n_field.name, 0)
        if dim == 2:
            start = ps.make_slice[L[0] // 2 - 1:L[0] // 2 + 1, L[1] // 2 - 1:L[1] // 2 + 1]
        else:
            start = ps.make_slice[L[0] // 2 - 1:L[0] // 2 + 1, L[1] // 2 - 1:L[1] // 2 + 1,
                                  L[2] // 2 - 1:L[2] // 2 + 1]
        dh.fill(n_field.name, 2**-dim, slice_obj=start)

        sync_conc()
        for i in range(time):
            dh.run_kernel(flux_kernel)
            dh.run_kernel(pde_kernel)
            sync_conc()

        sim_density = dh.gather_array(n_field.name)
        
        # check that mass was conserved
        assert np.isclose(sim_density.sum(), 1)
        assert np.all(sim_density > 0)
        
        # check that the maximum is in the right place
        peak = np.unravel_index(np.argmax(sim_density, axis=None), sim_density.shape)
        assert np.allclose(peak, np.array(L) // 2 - 0.5 + velocity * time, atol=0.5)
        
        # check the concentration profile
        if np.linalg.norm(velocity) == 0:
            calc_density = density(pos - velocity * time, time, D)
            target = [time, D]
        
            pytest.importorskip('scipy.optimize')
            from scipy.optimize import curve_fit
            popt, _ = curve_fit(lambda x, t, D: density(x - velocity * time, t, D),
                                pos.reshape(-1, dim),
                                sim_density.reshape(-1),
                                p0=target)
        
            assert np.isclose(popt[0], time, rtol=0.1)
            assert np.isclose(popt[1], D, rtol=0.1)
            assert np.allclose(calc_density, sim_density, atol=1e-4)

    return lambda v: run(np.array(v), time)


advection_diffusion.runners = {}


@pytest.mark.parametrize("velocity", list(product([0, -0.047, 0.041], [0, -0.031, 0.023])))
def test_advection_diffusion_2d(velocity):
    if 2 not in advection_diffusion.runners:
        advection_diffusion.runners[2] = advection_diffusion(2)
    advection_diffusion.runners[2](velocity)


@pytest.mark.parametrize("velocity", list(product([0, -0.047, 0.041], [0, -0.031, 0.023], [0, -0.017, 0.011])))
@pytest.mark.longrun
def test_advection_diffusion_3d(velocity):
    if 3 not in advection_diffusion.runners:
        advection_diffusion.runners[3] = advection_diffusion(3)
    advection_diffusion.runners[3](velocity)


def advection_diffusion_fluctuations(dim: int):
    # parameters
    if dim == 2:
        L = (32, 32)
        stencil_factor = np.sqrt(1 / (1 + np.sqrt(2)))
    elif dim == 3:
        L = (16, 16, 16)
        stencil_factor = np.sqrt(1 / (1 + 2 * np.sqrt(2) + 4.0 / 3.0 * np.sqrt(3)))

    dh = ps.create_data_handling(domain_size=L, periodicity=True, default_target=ps.Target.CPU)

    n_field = dh.add_array('n', values_per_cell=1)
    j_field = dh.add_array('j', values_per_cell=3 ** dim // 2, field_type=ps.FieldType.STAGGERED_FLUX)
    velocity_field = dh.add_array('v', values_per_cell=dim)

    D = 0.00666
    time = 10000

    def grad(f):
        return sp.Matrix([ps.fd.diff(f, i) for i in range(dim)])

    flux_eq = - D * grad(n_field)
    fvm_eq = ps.fd.FVM1stOrder(n_field, flux=flux_eq)

    vof_adv = ps.fd.VOF(j_field, velocity_field, n_field)

    # merge calculation of advection and diffusion terms
    flux = []
    for adv, div in zip(vof_adv, fvm_eq.discrete_flux(j_field)):
        assert adv.lhs == div.lhs
        flux.append(ps.Assignment(adv.lhs, adv.rhs + div.rhs))
    flux = ps.AssignmentCollection(flux)

    rng_symbol_gen = random_symbol(flux.subexpressions, dim=dh.dim)
    for i in range(len(flux.main_assignments)):
        n = j_field.staggered_stencil[i]
        assert flux.main_assignments[i].lhs == j_field.staggered_access(n)
        
        # calculate mean density
        dens = (n_field.neighbor_vector(n) + n_field.center_vector)[0] / 2
        # multyply by smoothed haviside function so that fluctuation will not get bigger that the density
        dens *= sp.Max(0, sp.Min(1.0, n_field.neighbor_vector(n)[0]) * sp.Min(1.0, n_field.center_vector[0]))
        
        # lenght of the vector
        length = sp.sqrt(len(j_field.staggered_stencil[i]))
        
        # amplitude of the random fluctuations
        fluct = sp.sqrt(2 * dens * D) * sp.sqrt(1 / length) * stencil_factor
        # add fluctuations
        fluct *= 2 * (next(rng_symbol_gen) - 0.5) * sp.sqrt(3)
        
        flux.main_assignments[i] = ps.Assignment(flux.main_assignments[i].lhs, flux.main_assignments[i].rhs + fluct)
    
    # Add the folding to the flux, so that the random numbers persist through the ghostlayers.
    fold = {ps.astnodes.LoopOverCoordinate.get_loop_counter_symbol(i):
            ps.astnodes.LoopOverCoordinate.get_loop_counter_symbol(i) % L[i] for i in range(len(L))}
    flux.subs(fold)

    flux_kernel = ps.create_staggered_kernel(flux).compile()

    pde_kernel = ps.create_kernel(fvm_eq.discrete_continuity(j_field)).compile()

    sync_conc = dh.synchronization_function([n_field.name])

    # analytical density distribution calculation
    def P(rho, density_init):
        res = []
        for r in rho:
            res.append(np.power(density_init, r) * np.exp(-density_init) / np.math.gamma(r + 1))
        return np.array(res)

    def run(density_init: float, velocity: np.ndarray, time: int):
        dh.fill(n_field.name, np.nan, ghost_layers=True, inner_ghost_layers=True)
        dh.fill(j_field.name, np.nan, ghost_layers=True, inner_ghost_layers=True)

        # set initial values for velocity and density
        for i in range(dim):
            dh.fill(velocity_field.name, velocity[i], i, ghost_layers=True, inner_ghost_layers=True)
        dh.fill(n_field.name, density_init)

        measurement_intervall = 10
        warm_up = 1000
        data = []

        sync_conc()
        for i in range(warm_up):
            dh.run_kernel(flux_kernel, seed=42, time_step=i)
            dh.run_kernel(pde_kernel)
            sync_conc()

        for i in range(time):
            dh.run_kernel(flux_kernel, seed=42, time_step=i + warm_up)
            dh.run_kernel(pde_kernel)
            sync_conc()
            if(i % measurement_intervall == 0):
                data = np.append(data, dh.gather_array(n_field.name).ravel(), 0)

        # test mass conservation
        np.testing.assert_almost_equal(dh.gather_array(n_field.name).mean(), density_init)

        n_bins = 50

        density_value, bins = np.histogram(data, density=True, bins=n_bins)
        bins_mean = bins[:-1] + (bins[1:] - bins[:-1]) / 2
        analytical_value = P(bins_mean, density_init)
        print(density_value - analytical_value)
        np.testing.assert_allclose(density_value, analytical_value, atol=2e-3)

    return lambda density_init, v: run(density_init, np.array(v), time)


advection_diffusion_fluctuations.runners = {}


@pytest.mark.parametrize("velocity", list(product([0, 0.00041], [0, -0.00031])))
@pytest.mark.parametrize("density", [27.0, 56.5])
@pytest.mark.longrun
def test_advection_diffusion_fluctuation_2d(density, velocity):
    if 2 not in advection_diffusion_fluctuations.runners:
        advection_diffusion_fluctuations.runners[2] = advection_diffusion_fluctuations(2)
    advection_diffusion_fluctuations.runners[2](density, velocity)


@pytest.mark.parametrize("velocity", [(0.0, 0.0, 0.0), (0.00043, -0.00017, 0.00028)])
@pytest.mark.parametrize("density", [27.0, 56.5])
@pytest.mark.longrun
def test_advection_diffusion_fluctuation_3d(density, velocity):
    if 3 not in advection_diffusion_fluctuations.runners:
        advection_diffusion_fluctuations.runners[3] = advection_diffusion_fluctuations(3)
    advection_diffusion_fluctuations.runners[3](density, velocity)


def diffusion_reaction(fluctuations: bool):
    # parameters
    L = (32, 32)
    stencil_factor = np.sqrt(1 / (1 + np.sqrt(2)))

    dh = ps.create_data_handling(domain_size=L, periodicity=True, default_target=ps.Target.CPU)

    species = 2
    n_fields = []
    j_fields = []
    r_flux_fields = []
    for i in range(species):
        n_fields.append(dh.add_array(f'n_{i}', values_per_cell=1))
        j_fields.append(dh.add_array(f'j_{i}', values_per_cell=3 ** dh.dim // 2,
                                     field_type=ps.FieldType.STAGGERED_FLUX))
        r_flux_fields.append(dh.add_array(f'r_{i}', values_per_cell=1))
    velocity_field = dh.add_array('v', values_per_cell=dh.dim)

    D = 0.00666
    time = 1000
    r_order = [2.0, 0.0]
    r_rate_const = 0.00001
    r_coefs = [-2, 1]

    def grad(f):
        return sp.Matrix([ps.fd.diff(f, i) for i in range(dh.dim)])

    flux_eq = - D * grad(n_fields[0])
    fvm_eq = ps.fd.FVM1stOrder(n_fields[0], flux=flux_eq)
    vof_adv = ps.fd.VOF(j_fields[0], velocity_field, n_fields[0])
    continuity_assignments = fvm_eq.discrete_continuity(j_fields[0])
    # merge calculation of advection and diffusion terms
    flux = []
    for adv, div in zip(vof_adv, fvm_eq.discrete_flux(j_fields[0])):
        assert adv.lhs == div.lhs
        flux.append(ps.Assignment(adv.lhs, adv.rhs + div.rhs))
    flux = ps.AssignmentCollection(flux)

    if(fluctuations):
        rng_symbol_gen = random_symbol(flux.subexpressions, dim=dh.dim)
        for i in range(len(flux.main_assignments)):
            n = j_fields[0].staggered_stencil[i]
            assert flux.main_assignments[i].lhs == j_fields[0].staggered_access(n)
            
            # calculate mean density
            dens = (n_fields[0].neighbor_vector(n) + n_fields[0].center_vector)[0] / 2
            # multyply by smoothed haviside function so that fluctuation will not get bigger that the density
            dens *= sp.Max(0,
                           sp.Min(1.0, n_fields[0].neighbor_vector(n)[0]) * sp.Min(1.0, n_fields[0].center_vector[0]))
            
            # lenght of the vector
            length = sp.sqrt(len(j_fields[0].staggered_stencil[i]))
            
            # amplitude of the random fluctuations
            fluct = sp.sqrt(2 * dens * D) * sp.sqrt(1 / length) * stencil_factor
            # add fluctuations
            fluct *= 2 * (next(rng_symbol_gen) - 0.5) * sp.sqrt(3)
            flux.main_assignments[i] = ps.Assignment(flux.main_assignments[i].lhs, flux.main_assignments[i].rhs + fluct)
        
        # Add the folding to the flux, so that the random numbers persist through the ghostlayers.
        fold = {ps.astnodes.LoopOverCoordinate.get_loop_counter_symbol(i):
                ps.astnodes.LoopOverCoordinate.get_loop_counter_symbol(i) % L[i] for i in range(len(L))}
        flux.subs(fold)

    r_flux = NodeCollection([SympyAssignment(j_fields[i].center, 0) for i in range(species)])
    reaction = r_rate_const
    for i in range(species):
        reaction *= sp.Pow(n_fields[i].center, r_order[i])
    new_assignments = []
    if fluctuations:
        rng_symbol_gen = random_symbol(new_assignments, dim=dh.dim)
        reaction_fluctuations = sp.sqrt(sp.Abs(reaction)) * 2 * (next(rng_symbol_gen) - 0.5) * sp.sqrt(3)
        reaction_fluctuations *= sp.Min(1, sp.Abs(reaction**2))
    else:
        reaction_fluctuations = 0.0
    for i in range(species):
        r_flux.all_assignments[i] = SympyAssignment(
            r_flux_fields[i].center, (reaction + reaction_fluctuations) * r_coefs[i])
    [r_flux.all_assignments.insert(0, new) for new in new_assignments]

    continuity_assignments = [SympyAssignment(*assignment.args) for assignment in continuity_assignments]
    continuity_assignments.append(SympyAssignment(n_fields[0].center, n_fields[0].center + r_flux_fields[0].center))

    flux_kernel = ps.create_staggered_kernel(flux).compile()
    reaction_kernel = ps.create_kernel(r_flux).compile()

    config = ps.CreateKernelConfig(allow_double_writes=True)
    pde_kernel = ps.create_kernel(continuity_assignments, config=config).compile()

    sync_conc = dh.synchronization_function([n_fields[0].name, n_fields[1].name])

    def f(t, r, n0, fac, fluctuations):
        """Calculates the amount of product created after a certain time of a reaction with form xA -> B

        Args:
            t: Time of the reation
            r: Reaction rate constant
            n0: Initial density of the 
            fac: Reaction order of A (this in most cases equals the stochometric coefficient x)
            fluctuations: Boolian whether fluctuations were included during the reaction.
        """
        if fluctuations:
            return 1 / fac * (n0 + n0 / (n0 - (n0 + 1) * np.exp(fac * r * t)))
        return 1 / fac * (n0 - (1 / (fac * r * t + (1 / n0))))

    def run(density_init: float, velocity: np.ndarray, time: int):
        for i in range(species):
            dh.fill(n_fields[i].name, np.nan, ghost_layers=True, inner_ghost_layers=True)
            dh.fill(j_fields[i].name, 0.0, ghost_layers=True, inner_ghost_layers=True)
            dh.fill(r_flux_fields[i].name, 0.0, ghost_layers=True, inner_ghost_layers=True)

        # set initial values for velocity and density
        for i in range(dh.dim):
            dh.fill(velocity_field.name, velocity[i], i, ghost_layers=True, inner_ghost_layers=True)
        dh.fill(n_fields[0].name, density_init)
        dh.fill(n_fields[1].name, 0.0)

        measurement_intervall = 10
        data = []
        
        sync_conc()
        for i in range(time):
            if(i % measurement_intervall == 0):
                data.append([i, dh.gather_array(n_fields[1].name).mean(), dh.gather_array(n_fields[0].name).mean()])
            dh.run_kernel(reaction_kernel, seed=41, time_step=i)
            for s_idx in range(species):
                flux_kernel(n_0=dh.cpu_arrays[n_fields[s_idx].name],
                            j_0=dh.cpu_arrays[j_fields[s_idx].name],
                            v=dh.cpu_arrays[velocity_field.name], seed=42 + s_idx, time_step=i)
                pde_kernel(n_0=dh.cpu_arrays[n_fields[s_idx].name],
                           j_0=dh.cpu_arrays[j_fields[s_idx].name],
                           r_0=dh.cpu_arrays[r_flux_fields[s_idx].name])
            sync_conc()

        data = np.array(data).transpose()
        x = data[0]
        analytical_value = f(x, r_rate_const, density_init, abs(r_coefs[0]), fluctuations)

        # test mass conservation
        np.testing.assert_almost_equal(
            dh.gather_array(n_fields[0].name).mean() + 2 * dh.gather_array(n_fields[1].name).mean(), density_init)

        r_tol = 2e-3
        if fluctuations:
            r_tol = 3e-2
        np.testing.assert_allclose(data[1], analytical_value, rtol=r_tol)

    return lambda density_init, v: run(density_init, np.array(v), time)


advection_diffusion_fluctuations.runners = {}


@pytest.mark.parametrize("velocity", list(product([0, 0.0041], [0, -0.0031])))
@pytest.mark.parametrize("density", [27.0, 56.5])
@pytest.mark.parametrize("fluctuations", [False, True])
@pytest.mark.longrun
def test_diffusion_reaction(fluctuations, density, velocity):
    diffusion_reaction.runner = diffusion_reaction(fluctuations)
    diffusion_reaction.runner(density, velocity)


def VOF2(j: ps.field.Field, v: ps.field.Field, ρ: ps.field.Field, simplify=True):
    """Volume-of-fluid discretization of advection

    Args:
        j: the staggered field to write the fluxes to. Should have a D2Q9/D3Q27 stencil. Other stencils work too, but
           incur a small error (D2Q5/D3Q7: v^2, D3Q19: v^3).
        v: the flow velocity field
        ρ: the quantity to advect
        simplify: whether to simplify the generated expressions (slow, but makes them much more readable and faster)
    """
    dim = j.spatial_dimensions
    assert ps.FieldType.is_staggered(j)
    
    def assume_velocity(e):
        if not simplify:
            return e
        repl = {}
        for c in e.atoms(sp.StrictGreaterThan, sp.GreaterThan):
            if isinstance(c.lhs, ps.field.Field.Access) and c.lhs.field == v and isinstance(c.rhs, sp.Number):
                if c.rhs <= -1:
                    repl[c] = True
                elif c.rhs >= 1:
                    repl[c] = False
        for c in e.atoms(sp.StrictLessThan, sp.LessThan):
            if isinstance(c.lhs, ps.field.Field.Access) and c.lhs.field == v and isinstance(c.rhs, sp.Number):
                if c.rhs >= 1:
                    repl[c] = True
                elif c.rhs <= -1:
                    repl[c] = False
        for c in e.atoms(sp.Equality):
            if isinstance(c.lhs, ps.field.Field.Access) and c.lhs.field == v and isinstance(c.rhs, sp.Number):
                if c.rhs <= -1 or c.rhs >= 1:
                    repl[c] = False
        return e.subs(repl)
    
    class AABB:
        def __init__(self, corner0, corner1):
            self.dim = len(corner0)
            self.minCorner = sp.zeros(self.dim, 1)
            self.maxCorner = sp.zeros(self.dim, 1)
            for i in range(self.dim):
                self.minCorner[i] = sp.Piecewise((corner0[i], corner0[i] < corner1[i]), (corner1[i], True))
                self.maxCorner[i] = sp.Piecewise((corner1[i], corner0[i] < corner1[i]), (corner0[i], True))

        def intersect(self, other):
            minCorner = [sp.Max(self.minCorner[d], other.minCorner[d]) for d in range(self.dim)]
            maxCorner = [sp.Max(minCorner[d], sp.Min(self.maxCorner[d], other.maxCorner[d]))
                         for d in range(self.dim)]
            return AABB(minCorner, maxCorner)

        @property
        def volume(self):
            v = sp.prod([self.maxCorner[d] - self.minCorner[d] for d in range(self.dim)])
            if simplify:
                return sp.simplify(assume_velocity(v.rewrite(sp.Piecewise)))
            else:
                return v
    
    fluxes = []
    cell = AABB([-0.5] * dim, [0.5] * dim)
    cell_s = AABB(sp.Matrix([-0.5] * dim) + v.center_vector, sp.Matrix([0.5] * dim) + v.center_vector)
    for d, neighbor in enumerate(j.staggered_stencil):
        c = sp.Matrix(ps.stencil.direction_string_to_offset(neighbor)[:dim])
        cell_n = AABB(sp.Matrix([-0.5] * dim) + c, sp.Matrix([0.5] * dim) + c)
        cell_ns = AABB(sp.Matrix([-0.5] * dim) + c + v.neighbor_vector(neighbor),
                       sp.Matrix([0.5] * dim) + c + v.neighbor_vector(neighbor))
        fluxes.append(assume_velocity(ρ.center_vector * cell_s.intersect(cell_n).volume
                                      - ρ.neighbor_vector(neighbor) * cell_ns.intersect(cell).volume))
    
    assignments = []
    for i, d in enumerate(j.staggered_stencil):
        for lhs, rhs in zip(j.staggered_vector_access(d).values(), fluxes[i].values()):
            assignments.append(ps.Assignment(lhs, rhs))
    return assignments


@pytest.mark.parametrize("dim", [2, 3])
def test_advection(dim):
    L = (8,) * dim
    dh = ps.create_data_handling(L, periodicity=True, default_target=ps.Target.CPU)
    c = dh.add_array('c', values_per_cell=1)
    j = dh.add_array('j', values_per_cell=3 ** dh.dim // 2, field_type=ps.FieldType.STAGGERED_FLUX)
    u = dh.add_array('u', values_per_cell=dh.dim)
    
    dh.cpu_arrays[c.name][:] = (np.random.random([l + 2 for l in L]))
    dh.cpu_arrays[u.name][:] = (np.random.random([l + 2 for l in L] + [dim]) - 0.5) / 5
    
    vof1 = ps.create_kernel(ps.fd.VOF(j, u, c)).compile()
    dh.fill(j.name, np.nan, ghost_layers=True)
    dh.run_kernel(vof1)
    j1 = dh.gather_array(j.name).copy()
    
    vof2 = ps.create_kernel(VOF2(j, u, c, simplify=False)).compile()
    dh.fill(j.name, np.nan, ghost_layers=True)
    dh.run_kernel(vof2)
    j2 = dh.gather_array(j.name)
    
    assert np.allclose(j1, j2)


@pytest.mark.parametrize("stencil", ["D2Q5", "D2Q9"])
def test_ek(stencil):

    # parameters

    L = (40, 40)
    D = sp.Symbol("D")
    z = sp.Symbol("z")

    # data structures

    dh = ps.create_data_handling(L, periodicity=True, default_target=ps.Target.CPU)
    c = dh.add_array('c', values_per_cell=1)
    j = dh.add_array('j', values_per_cell=int(stencil[-1]) // 2, field_type=ps.FieldType.STAGGERED_FLUX)
    Phi = dh.add_array('Φ', values_per_cell=1)

    # perform automatic discretization

    def Gradient(f):
        return sp.Matrix([ps.fd.diff(f, i) for i in range(dh.dim)])

    flux_eq = -D * Gradient(c) + D * z * c.center * Gradient(Phi)

    disc = ps.fd.FVM1stOrder(c, flux_eq)
    flux_assignments = disc.discrete_flux(j)
    continuity_assignments = disc.discrete_continuity(j)

    # manual discretization

    x_staggered = - c[-1, 0] + c[0, 0] + z * (c[-1, 0] + c[0, 0]) / 2 * (Phi[-1, 0] - Phi[0, 0])
    y_staggered = - c[0, -1] + c[0, 0] + z * (c[0, -1] + c[0, 0]) / 2 * (Phi[0, -1] - Phi[0, 0])
    xy_staggered = (- c[-1, -1] + c[0, 0]) / sp.sqrt(2) + \
        z * (c[-1, -1] + c[0, 0]) / 2 * (Phi[-1, -1] - Phi[0, 0]) / sp.sqrt(2)
    xY_staggered = (- c[-1, 1] + c[0, 0]) / sp.sqrt(2) + \
        z * (c[-1, 1] + c[0, 0]) / 2 * (Phi[-1, 1] - Phi[0, 0]) / sp.sqrt(2)
    A0 = (1 + sp.sqrt(2) if j.index_shape[0] == 4 else 1)

    jj = j.staggered_access
    divergence = -1 * sum([jj(d) for d in j.staggered_stencil
                          + [ps.stencil.inverse_direction_string(d) for d in j.staggered_stencil]])

    update = [ps.Assignment(c.center, c.center + divergence)]
    flux = [ps.Assignment(j.staggered_access("W"), D * x_staggered / A0),
            ps.Assignment(j.staggered_access("S"), D * y_staggered / A0)]
    if j.index_shape[0] == 4:
        flux += [ps.Assignment(j.staggered_access("SW"), D * xy_staggered / A0),
                 ps.Assignment(j.staggered_access("NW"), D * xY_staggered / A0)]

    # compare

    for a, b in zip(flux, flux_assignments):
        assert a.lhs == b.lhs
        assert sp.simplify(a.rhs - b.rhs) == 0
    for a, b in zip(update, continuity_assignments):
        assert a.lhs == b.lhs
        assert a.rhs == b.rhs

# TODO: test source


@pytest.mark.parametrize("stencil", ["D2Q5", "D2Q9", "D3Q7", "D3Q19", "D3Q27"])
@pytest.mark.parametrize("derivative", [0, 1])
def test_flux_stencil(stencil, derivative):
    L = (40, ) * int(stencil[1])
    dh = ps.create_data_handling(L, periodicity=True, default_target=ps.Target.CPU)
    c = dh.add_array('c', values_per_cell=1)
    j = dh.add_array('j', values_per_cell=int(stencil[3:]) // 2, field_type=ps.FieldType.STAGGERED_FLUX)

    def Gradient(f):
        return sp.Matrix([ps.fd.diff(f, i) for i in range(dh.dim)])

    eq = [sp.Matrix([sp.Symbol(f"a_{i}") * c.center for i in range(dh.dim)]), Gradient(c)][derivative]
    disc = ps.fd.FVM1stOrder(c, flux=eq)

    # check the continuity
    continuity_assignments = disc.discrete_continuity(j)
    assert [len(a.rhs.atoms(ps.field.Field.Access)) for a in continuity_assignments] == \
           [int(stencil[3:])] * len(continuity_assignments)

    # check the flux
    flux_assignments = disc.discrete_flux(j)
    assert [len(a.rhs.atoms(ps.field.Field.Access)) for a in flux_assignments] == [2] * len(flux_assignments)


@pytest.mark.parametrize("stencil", ["D2Q5", "D2Q9", "D3Q7", "D3Q19", "D3Q27"])
def test_source_stencil(stencil):
    L = (40, ) * int(stencil[1])
    dh = ps.create_data_handling(L, periodicity=True, default_target=ps.Target.CPU)
    c = dh.add_array('c', values_per_cell=1)
    j = dh.add_array('j', values_per_cell=int(stencil[3:]) // 2, field_type=ps.FieldType.STAGGERED_FLUX)

    continuity_ref = ps.fd.FVM1stOrder(c).discrete_continuity(j)

    for eq in [c.center] + [ps.fd.diff(c, i) for i in range(dh.dim)]:
        disc = ps.fd.FVM1stOrder(c, source=eq)
        diff = sp.simplify(disc.discrete_continuity(j)[0].rhs - continuity_ref[0].rhs)
        if type(eq) is ps.field.Field.Access:
            assert len(diff.atoms(ps.field.Field.Access)) == 1
        else:
            assert len(diff.atoms(ps.field.Field.Access)) == 2


def test_fvm_staggered_simplification():
    D = sp.Symbol("D")
    data_type = "float64"

    c = ps.fields(f"c: {data_type}[2D]", layout='fzyx')
    j = ps.fields(f"j(2): {data_type}[2D]", layout='fzyx', field_type=ps.FieldType.STAGGERED_FLUX)

    grad_c = sp.Matrix([ps.fd.diff(c, i) for i in range(c.spatial_dimensions)])

    ek = ps.fd.FVM1stOrder(c, flux=-D * grad_c)

    ast = ps.create_staggered_kernel(ek.discrete_flux(j))
    code = ps.get_code_str(ast)

    assert '_size_c_0 - 1 < _size_c_0 - 1' not in code



import numpy as np
import pycuda.autoinit
import pycuda.gpuarray as gpuarray
import sympy as sp
from scipy.ndimage import convolve

from pystencils import Assignment, Field, fields, CreateKernelConfig, create_kernel, Target
from pystencils.gpucuda import BlockIndexing
from pystencils.simp import sympy_cse_on_assignment_list
from pystencils.slicing import add_ghost_layers, make_slice, remove_ghost_layers


def test_averaging_kernel():
    size = (40, 55)
    src_arr = np.random.rand(*size)
    src_arr = add_ghost_layers(src_arr)
    dst_arr = np.zeros_like(src_arr)
    src_field = Field.create_from_numpy_array('src', src_arr)
    dst_field = Field.create_from_numpy_array('dst', dst_arr)

    update_rule = Assignment(dst_field[0, 0],
                             (src_field[0, 1] + src_field[0, -1] + src_field[1, 0] + src_field[-1, 0]) / 4)

    config = CreateKernelConfig(target=Target.GPU)
    ast = create_kernel(sympy_cse_on_assignment_list([update_rule]), config=config)
    kernel = ast.compile()

    gpu_src_arr = gpuarray.to_gpu(src_arr)
    gpu_dst_arr = gpuarray.to_gpu(dst_arr)
    kernel(src=gpu_src_arr, dst=gpu_dst_arr)
    gpu_dst_arr.get(dst_arr)

    stencil = np.array([[0, 1, 0], [1, 0, 1], [0, 1, 0]]) / 4.0
    reference = convolve(remove_ghost_layers(src_arr), stencil, mode='constant', cval=0.0)
    reference = add_ghost_layers(reference)
    np.testing.assert_almost_equal(reference, dst_arr)


def test_variable_sized_fields():
    src_field = Field.create_generic('src', spatial_dimensions=2)
    dst_field = Field.create_generic('dst', spatial_dimensions=2)

    update_rule = Assignment(dst_field[0, 0],
                             (src_field[0, 1] + src_field[0, -1] + src_field[1, 0] + src_field[-1, 0]) / 4)

    config = CreateKernelConfig(target=Target.GPU)
    ast = create_kernel(sympy_cse_on_assignment_list([update_rule]), config=config)
    kernel = ast.compile()

    size = (3, 3)
    src_arr = np.random.rand(*size)
    src_arr = add_ghost_layers(src_arr)
    dst_arr = np.zeros_like(src_arr)

    gpu_src_arr = gpuarray.to_gpu(src_arr)
    gpu_dst_arr = gpuarray.to_gpu(dst_arr)
    kernel(src=gpu_src_arr, dst=gpu_dst_arr)
    gpu_dst_arr.get(dst_arr)

    stencil = np.array([[0, 1, 0], [1, 0, 1], [0, 1, 0]]) / 4.0
    reference = convolve(remove_ghost_layers(src_arr), stencil, mode='constant', cval=0.0)
    reference = add_ghost_layers(reference)
    np.testing.assert_almost_equal(reference, dst_arr)


def test_multiple_index_dimensions():
    """Sums along the last axis of a numpy array"""
    src_size = (7, 6, 4)
    dst_size = src_size[:2]
    src_arr = np.asfortranarray(np.random.rand(*src_size))
    dst_arr = np.zeros(dst_size)

    src_field = Field.create_from_numpy_array('src', src_arr, index_dimensions=1)
    dst_field = Field.create_from_numpy_array('dst', dst_arr, index_dimensions=0)

    offset = (-2, -1)
    update_rule = Assignment(dst_field[0, 0],
                             sum([src_field[offset[0], offset[1]](i) for i in range(src_size[-1])]))

    config = CreateKernelConfig(target=Target.GPU)
    ast = create_kernel([update_rule], config=config)
    kernel = ast.compile()

    gpu_src_arr = gpuarray.to_gpu(src_arr)
    gpu_dst_arr = gpuarray.to_gpu(dst_arr)
    kernel(src=gpu_src_arr, dst=gpu_dst_arr)
    gpu_dst_arr.get(dst_arr)

    reference = np.zeros_like(dst_arr)
    gl = np.max(np.abs(np.array(offset, dtype=int)))
    for x in range(gl, src_size[0]-gl):
        for y in range(gl, src_size[1]-gl):
            reference[x, y] = sum([src_arr[x+offset[0], y+offset[1], i] for i in range(src_size[2])])

    np.testing.assert_almost_equal(reference, dst_arr)


def test_ghost_layer():
    size = (6, 5)
    src_arr = np.ones(size)
    dst_arr = np.zeros_like(src_arr)
    src_field = Field.create_from_numpy_array('src', src_arr, index_dimensions=0)
    dst_field = Field.create_from_numpy_array('dst', dst_arr, index_dimensions=0)

    update_rule = Assignment(dst_field[0, 0], src_field[0, 0])
    ghost_layers = [(1, 2), (2, 1)]

    config = CreateKernelConfig(target=Target.GPU, ghost_layers=ghost_layers, gpu_indexing="line")
    ast = create_kernel(sympy_cse_on_assignment_list([update_rule]), config=config)
    kernel = ast.compile()

    gpu_src_arr = gpuarray.to_gpu(src_arr)
    gpu_dst_arr = gpuarray.to_gpu(dst_arr)
    kernel(src=gpu_src_arr, dst=gpu_dst_arr)
    gpu_dst_arr.get(dst_arr)

    reference = np.zeros_like(src_arr)
    reference[ghost_layers[0][0]:-ghost_layers[0][1], ghost_layers[1][0]:-ghost_layers[1][1]] = 1
    np.testing.assert_equal(reference, dst_arr)


def test_setting_value():
    arr_cpu = np.arange(25, dtype=np.float64).reshape(5, 5)
    arr_gpu = gpuarray.to_gpu(arr_cpu)

    iteration_slice = make_slice[:, :]
    f = Field.create_generic("f", 2)
    update_rule = [Assignment(f(0), sp.Symbol("value"))]

    config = CreateKernelConfig(target=Target.GPU, gpu_indexing="line", iteration_slice=iteration_slice)
    ast = create_kernel(sympy_cse_on_assignment_list(update_rule), config=config)
    kernel = ast.compile()

    kernel(f=arr_gpu, value=np.float64(42.0))
    np.testing.assert_equal(arr_gpu.get(), np.ones((5, 5)) * 42.0)


def test_periodicity():
    from pystencils.gpucuda.periodicity import get_periodic_boundary_functor as periodic_gpu
    from pystencils.slicing import get_periodic_boundary_functor as periodic_cpu

    arr_cpu = np.arange(50, dtype=np.float64).reshape(5, 5, 2)
    arr_gpu = gpuarray.to_gpu(arr_cpu)

    periodicity_stencil = [(1, 0), (-1, 0), (1, 1)]
    periodic_gpu_kernel = periodic_gpu(periodicity_stencil, (5, 5), 1, 2)
    periodic_cpu_kernel = periodic_cpu(periodicity_stencil)

    cpu_result = np.copy(arr_cpu)
    periodic_cpu_kernel(cpu_result)

    gpu_result = np.copy(arr_cpu)
    periodic_gpu_kernel(pdfs=arr_gpu)
    arr_gpu.get(gpu_result)
    np.testing.assert_equal(cpu_result, gpu_result)


def test_block_indexing():
    f = fields("f: [3D]")
    bi = BlockIndexing(f, make_slice[:, :, :], block_size=(16, 8, 2), permute_block_size_dependent_on_layout=False)
    assert bi.call_parameters((3, 2, 32))['block'] == (3, 2, 32)
    assert bi.call_parameters((32, 2, 32))['block'] == (16, 2, 8)

    bi = BlockIndexing(f, make_slice[:, :, :], block_size=(32, 1, 1), permute_block_size_dependent_on_layout=False)
    assert bi.call_parameters((1, 16, 16))['block'] == (1, 16, 2)

    bi = BlockIndexing(f, make_slice[:, :, :], block_size=(16, 8, 2), maximum_block_size="auto")
    # This function should be used if number of needed registers is known. Can be determined with func.num_regs
    blocks = bi.limit_block_size_by_register_restriction([1024, 1024, 1], 1000)

    assert sum(blocks) < sum([1024, 1024, 1])



from itertools import chain
from typing import Callable, List, Sequence, Union
from collections import defaultdict

import sympy as sp

from pystencils.assignment import Assignment
from pystencils.astnodes import Node
from pystencils.field import Field
from pystencils.sympyextensions import subs_additive, is_constant, recursive_collect


def sort_assignments_topologically(assignments: Sequence[Union[Assignment, Node]]) -> List[Union[Assignment, Node]]:
    """Sorts assignments in topological order, such that symbols used on rhs occur first on a lhs"""
    edges = []
    for c1, e1 in enumerate(assignments):
        if hasattr(e1, 'lhs') and hasattr(e1, 'rhs'):
            symbols = [e1.lhs]
        elif isinstance(e1, Node):
            symbols = e1.symbols_defined
        else:
            raise NotImplementedError(f"Cannot sort topologically. Object of type {type(e1)} cannot be handled.")

        for lhs in symbols:
            for c2, e2 in enumerate(assignments):
                if isinstance(e2, Assignment) and lhs in e2.rhs.free_symbols:
                    edges.append((c1, c2))
                elif isinstance(e2, Node) and lhs in e2.undefined_symbols:
                    edges.append((c1, c2))
    return [assignments[i] for i in sp.topological_sort((range(len(assignments)), edges))]


def sympy_cse(ac, **kwargs):
    """Searches for common subexpressions inside the assignment collection.

    Searches is done in both the existing subexpressions as well as the assignments themselves.
    It uses the sympy subexpression detection to do this. Return a new assignment collection
    with the additional subexpressions found
    """
    symbol_gen = ac.subexpression_symbol_generator

    all_assignments = [e for e in chain(ac.subexpressions, ac.main_assignments) if isinstance(e, Assignment)]
    other_objects = [e for e in chain(ac.subexpressions, ac.main_assignments) if not isinstance(e, Assignment)]
    replacements, new_eq = sp.cse(all_assignments, symbols=symbol_gen, **kwargs)

    replacement_eqs = [Assignment(*r) for r in replacements]

    modified_subexpressions = new_eq[:len(ac.subexpressions)]
    modified_update_equations = new_eq[len(ac.subexpressions):]

    new_subexpressions = sort_assignments_topologically(other_objects + replacement_eqs + modified_subexpressions)
    return ac.copy(modified_update_equations, new_subexpressions)


def sympy_cse_on_assignment_list(assignments: List[Assignment]) -> List[Assignment]:
    """Extracts common subexpressions from a list of assignments."""
    from pystencils.simp.assignment_collection import AssignmentCollection
    ec = AssignmentCollection([], assignments)
    return sympy_cse(ec).all_assignments


def subexpression_substitution_in_existing_subexpressions(ac):
    """Goes through the subexpressions list and replaces the term in the following subexpressions."""
    result = []
    for outer_ctr, s in enumerate(ac.subexpressions):
        new_rhs = s.rhs
        for inner_ctr in range(outer_ctr):
            sub_expr = ac.subexpressions[inner_ctr]
            new_rhs = subs_additive(new_rhs, sub_expr.lhs, sub_expr.rhs, required_match_replacement=1.0)
            new_rhs = new_rhs.subs(sub_expr.rhs, sub_expr.lhs)
        result.append(Assignment(s.lhs, new_rhs))

    return ac.copy(ac.main_assignments, result)


def subexpression_substitution_in_main_assignments(ac):
    """Replaces already existing subexpressions in the equations of the assignment_collection."""
    result = []
    for s in ac.main_assignments:
        new_rhs = s.rhs
        for sub_expr in ac.subexpressions:
            new_rhs = subs_additive(new_rhs, sub_expr.lhs, sub_expr.rhs, required_match_replacement=1.0)
        result.append(Assignment(s.lhs, new_rhs))
    return ac.copy(result)


def add_subexpressions_for_constants(ac):
    """Extracts constant factors to subexpressions in the given assignment collection.

    SymPy will exclude common factors from a sum only if they are symbols. This simplification
    can be applied to exclude common numeric constants from multiple terms of a sum. As a consequence,
    the number of multiplications is reduced and in some cases, more common subexpressions can be found.
    """
    constants_to_subexp_dict = defaultdict(lambda: next(ac.subexpression_symbol_generator))

    def visit(expr):
        args = list(expr.args)
        if len(args) == 0:
            return expr
        if isinstance(expr, sp.Add) or isinstance(expr, sp.Mul):
            for i, arg in enumerate(args):
                if is_constant(arg) and abs(arg) != 1:
                    if arg < 0:
                        args[i] = - constants_to_subexp_dict[- arg]
                    else:
                        args[i] = constants_to_subexp_dict[arg]
        return expr.func(*(visit(a) for a in args))
    main_assignments = [Assignment(a.lhs, visit(a.rhs)) for a in ac.main_assignments]
    subexpressions = [Assignment(a.lhs, visit(a.rhs)) for a in ac.subexpressions]

    symbols_to_collect = set(constants_to_subexp_dict.values())

    main_assignments = [Assignment(a.lhs, recursive_collect(a.rhs, symbols_to_collect, True)) for a in main_assignments]
    subexpressions = [Assignment(a.lhs, recursive_collect(a.rhs, symbols_to_collect, True)) for a in subexpressions]

    subexpressions = [Assignment(symb, c) for c, symb in constants_to_subexp_dict.items()] + subexpressions
    return ac.copy(main_assignments=main_assignments, subexpressions=subexpressions)


def add_subexpressions_for_divisions(ac):
    r"""Introduces subexpressions for all divisions which have no constant in the denominator.

    For example :math:`\frac{1}{x}` is replaced while :math:`\frac{1}{3}` is not replaced.
    """
    divisors = set()

    def search_divisors(term):
        if term.func == sp.Pow:
            if term.exp.is_integer and term.exp.is_number and term.exp < 0:
                divisors.add(term)
        else:
            for a in term.args:
                search_divisors(a)

    for eq in ac.all_assignments:
        search_divisors(eq.rhs)

    divisors = sorted(list(divisors), key=lambda x: str(x))
    new_symbol_gen = ac.subexpression_symbol_generator
    substitutions = {divisor: new_symbol for new_symbol, divisor in zip(new_symbol_gen, divisors)}
    return ac.new_with_substitutions(substitutions, add_substitutions_as_subexpressions=True, substitute_on_lhs=False)


def add_subexpressions_for_sums(ac):
    r"""Introduces subexpressions for all sums - i.e. splits addends into subexpressions."""
    addends = []

    def contains_sum(term):
        if term.func == sp.Add:
            return True
        if term.is_Atom:
            return False
        return any([contains_sum(a) for a in term.args])

    def search_addends(term):
        if term.func == sp.Add:
            if all([not contains_sum(a) for a in term.args]):
                addends.extend(term.args)
        for a in term.args:
            search_addends(a)

    for eq in ac.all_assignments:
        search_addends(eq.rhs)

    addends = [a for a in addends if not isinstance(a, sp.Symbol) or isinstance(a, Field.Access)]
    new_symbol_gen = ac.subexpression_symbol_generator
    substitutions = {addend: new_symbol for new_symbol, addend in zip(new_symbol_gen, addends)}
    return ac.new_with_substitutions(substitutions, True, substitute_on_lhs=False)


def add_subexpressions_for_field_reads(ac, subexpressions=True, main_assignments=True):
    r"""Substitutes field accesses on rhs of assignments with subexpressions

    Can change semantics of the update rule (which is the goal of this transformation)
    This is useful if a field should be update in place - all values are loaded before into subexpression variables,
    then the new values are computed and written to the same field in-place.
    """
    field_reads = set()
    to_iterate = []
    if subexpressions:
        to_iterate = chain(to_iterate, ac.subexpressions)
    if main_assignments:
        to_iterate = chain(to_iterate, ac.main_assignments)

    for assignment in to_iterate:
        if hasattr(assignment, 'lhs') and hasattr(assignment, 'rhs'):
            field_reads.update(assignment.rhs.atoms(Field.Access))
    substitutions = {fa: next(ac.subexpression_symbol_generator) for fa in field_reads}
    return ac.new_with_substitutions(substitutions, add_substitutions_as_subexpressions=True,
                                     substitute_on_lhs=False, sort_topologically=False)


def transform_rhs(assignment_list, transformation, *args, **kwargs):
    """Applies a transformation function on the rhs of each element of the passed assignment list
    If the list also contains other object, like AST nodes, these are ignored.
    Additional parameters are passed to the transformation function"""
    return [Assignment(a.lhs, transformation(a.rhs, *args, **kwargs)) if hasattr(a, 'lhs') and hasattr(a, 'rhs') else a
            for a in assignment_list]


def transform_lhs_and_rhs(assignment_list, transformation, *args, **kwargs):
    return [Assignment(transformation(a.lhs, *args, **kwargs),
                       transformation(a.rhs, *args, **kwargs))
            if hasattr(a, 'lhs') and hasattr(a, 'rhs') else a
            for a in assignment_list]


def apply_to_all_assignments(operation: Callable[[sp.Expr], sp.Expr]):
    """Applies a given operation to all equations in collection."""

    def f(ac):
        return ac.copy(transform_rhs(ac.main_assignments, operation))

    f.__name__ = operation.__name__
    return f


def apply_on_all_subexpressions(operation: Callable[[sp.Expr], sp.Expr]):
    """Applies the given operation on all subexpressions of the AC."""

    def f(ac):
        return ac.copy(ac.main_assignments, transform_rhs(ac.subexpressions, operation))

    f.__name__ = operation.__name__
    return f

# TODO Markus
# make this really work for Assignmentcollections
# this function should ONLY evaluate
# do the optims_c99 elsewhere optionally

# def apply_sympy_optimisations(ac: AssignmentCollection):
#     """ Evaluates constant expressions (e.g. :math:`\\sqrt{3}` will be replaced by its floating point representation)
#         and applies the default sympy optimisations. See sympy.codegen.rewriting
#     """
#
#     # Evaluates all constant terms
#
#     assignments = ac.all_assignments
#
#     evaluate_constant_terms = ReplaceOptim(lambda e: hasattr(e, 'is_constant') and e.is_constant and not e.is_integer,
#                                            lambda p: p.evalf())
#
#     sympy_optimisations = [evaluate_constant_terms] + list(optims_c99)
#
#     assignments = [Assignment(a.lhs, optimize(a.rhs, sympy_optimisations))
#                    if hasattr(a, 'lhs')
#                    else a for a in assignments]
#     assignments_nodes = [a.atoms(SympyAssignment) for a in assignments]
#     for a in chain.from_iterable(assignments_nodes):
#         a.optimize(sympy_optimisations)
#
#     return AssignmentCollection(assignments)



# -*- coding: utf-8 -*-
#
# Copyright © 2019 Stephan Seitz <stephan.seitz@fau.de>
#
# Distributed under terms of the GPLv3 license.

"""

"""

import sympy as sp

import pystencils
from pystencils.typing import TypedSymbol, BasicType


def test_wild_typed_symbol():
    x = pystencils.fields('x:  float32[3d]')
    typed_symbol = TypedSymbol('a', BasicType('float64'))

    assert x.center().match(sp.Wild('w1'))
    assert typed_symbol.match(sp.Wild('w1'))

    wild_ceiling = sp.ceiling(sp.Wild('w1'))
    assert sp.ceiling(x.center()).match(wild_ceiling)
    assert sp.ceiling(typed_symbol).match(wild_ceiling)


def test_replace_and_subs_for_assignment_collection():

    x, y = pystencils.fields('x, y:  float32[3d]')
    a, b, c, d = sp.symbols('a, b, c, d')

    assignments = pystencils.AssignmentCollection({
        a: sp.floor(1),
        b: 2,
        c: a + c,
        y.center(): sp.ceiling(x.center()) + sp.floor(x.center())
    })

    expected_assignments = pystencils.AssignmentCollection({
        a: sp.floor(3),
        b: 2,
        c: a + c,
        y.center(): sp.ceiling(x.center()) + sp.floor(x.center())
    })

    assert expected_assignments == assignments.replace(1, 3)
    assert expected_assignments == assignments.subs({1: 3})

    expected_assignments = pystencils.AssignmentCollection({
        d: sp.floor(1),
        b: 2,
        c: d + c,
        y.center(): sp.ceiling(x.center()) + sp.floor(x.center())
    })

    print(expected_assignments)
    print(assignments.subs(a, d))
    assert expected_assignments == assignments.subs(a, d)


def test_match_for_assignment_collection():

    x, y = pystencils.fields('x, y:  float32[3d]')
    a, b, c, d = sp.symbols('a, b, c, d')

    assignments = pystencils.AssignmentCollection({
        a: sp.floor(1),
        b: 2,
        c: a + c,
        y.center(): sp.ceiling(x.center()) + sp.floor(x.center())
    })

    w1 = sp.Wild('w1')
    w2 = sp.Wild('w2')
    w3 = sp.Wild('w3')

    wild_ceiling = sp.ceiling(w1)
    wild_addition = w1 + w2

    assert assignments.match(pystencils.Assignment(w3, wild_ceiling + w2))[w1] == x.center()
    assert assignments.match(pystencils.Assignment(w3, wild_ceiling + w2)) == {
        w3: y.center(),
        w2: sp.floor(x.center()),
        w1: x.center()
    }
    assert assignments.find(wild_ceiling) == {sp.ceiling(x.center())}
    assert len([a for a in assignments.find(wild_addition) if isinstance(a, sp.Add)]) == 2



# -*- coding: utf-8 -*-
#
# Copyright © 2019 Stephan Seitz <stephan.seitz@fau.de>
#
# Distributed under terms of the GPLv3 license.

"""

"""
import pytest
import numpy as np

import pystencils.config
import sympy as sp
import sympy.abc

import pystencils as ps
from pystencils.typing import create_type


@pytest.mark.parametrize('dtype', ["float64", "float32"])
def test_sum(dtype):

    sum = sp.Sum(sp.abc.k, (sp.abc.k, 1, 100))
    expanded_sum = sum.doit()

    # print(sum)
    # print(expanded_sum)

    x = ps.fields(f'x: {dtype}[1d]')

    assignments = ps.AssignmentCollection({x.center(): sum})

    ast = ps.create_kernel(assignments)
    code = ps.get_code_str(ast)
    kernel = ast.compile()

    # ps.show_code(ast)

    if dtype == "float32":
        assert "5050.0f;" in code

    array = np.zeros((10,), dtype=dtype)
    kernel(x=array)
    assert np.allclose(array, int(expanded_sum) * np.ones_like(array))


@pytest.mark.parametrize('dtype', ["int32", "int64", "float64", "float32"])
def test_product(dtype):

    k = ps.TypedSymbol('k', create_type(dtype))

    sum = sympy.Product(k, (k, 1, 10))
    expanded_sum = sum.doit()

    # print(sum)
    # print(expanded_sum)

    x = ps.fields(f'x: {dtype}[1d]')

    assignments = ps.AssignmentCollection({x.center(): sum})

    config = pystencils.config.CreateKernelConfig()

    ast = ps.create_kernel(assignments, config=config)
    code = ps.get_code_str(ast)
    kernel = ast.compile()

    # print(code)
    if dtype == "int64" or dtype == "int32":
        assert '3628800;' in code
    elif dtype == "float32":
        assert '3628800.0f;' in code
    else:
        assert '3628800.0;' in code

    array = np.zeros((10,), dtype=dtype)
    kernel(x=array)
    assert np.allclose(array, int(expanded_sum) * np.ones_like(array))

# TODO: See Issue !55
# def test_prod_var_limit():
#
#     k = ps.TypedSymbol('k', create_type('int64'))
#     limit = ps.TypedSymbol('limit', create_type('int64'))
#
#     sum = sympy.Sum(k, (k, 1, limit))
#     expanded_sum = sum.replace(limit, 100).doit()
#
#     print(sum)
#     print(expanded_sum)
#
#     x = ps.fields('x: int64[1d]')
#
#     assignments = ps.AssignmentCollection({x.center(): sum})
#
#     ast = ps.create_kernel(assignments)
#     ps.show_code(ast)
#     kernel = ast.compile()
#
#     array = np.zeros((10,), np.int64)
#
#     kernel(x=array, limit=100)
#
#     assert np.allclose(array, int(expanded_sum) * np.ones_like(array))



import numpy as np
from itertools import product
import pystencils.boundaries.createindexlist as cil

import pytest


@pytest.mark.parametrize('single_link', [False, True])
@pytest.mark.skipif(not cil.cython_funcs_available, reason='Cython functions are not available')
def test_equivalence_cython_python_version(single_link):
    #   D2Q9
    stencil_2d = tuple((x, y) for x, y in product([-1, 0, 1], [-1, 0, 1]))
    #   D3Q19
    stencil_3d = tuple(
        (x, y, z) for x, y, z in product([-1, 0, 1], [-1, 0, 1], [-1, 0, 1]) if abs(x) + abs(y) + abs(z) < 3)

    for dtype in [int, np.int16, np.uint32]:
        fluid_mask = dtype(1)
        mask = dtype(2)
        flag_field_2d = np.ones([15, 16], dtype=dtype) * fluid_mask
        flag_field_3d = np.ones([15, 16, 17], dtype=dtype) * fluid_mask

        flag_field_2d[0, :] = mask
        flag_field_2d[-1, :] = mask
        flag_field_2d[7, 7] = mask

        flag_field_3d[0, :, :] = mask
        flag_field_3d[-1, :, :] = mask
        flag_field_3d[7, 7, 7] = mask

        result_python_2d = cil._create_index_list_python(flag_field_2d, mask, fluid_mask,
                                                         stencil_2d, single_link, True, 1)

        result_python_3d = cil._create_index_list_python(flag_field_3d, mask, fluid_mask,
                                                         stencil_3d, single_link, True, 1)

        result_cython_2d = cil.create_boundary_index_list(flag_field_2d, stencil_2d, mask,
                                                          fluid_mask, 1, True, single_link)
        result_cython_3d = cil.create_boundary_index_list(flag_field_3d, stencil_3d, mask,
                                                          fluid_mask, 1, True, single_link)

        np.testing.assert_equal(result_python_2d, result_cython_2d)
        np.testing.assert_equal(result_python_3d, result_cython_3d)


@pytest.mark.parametrize('single_link', [False, True])
@pytest.mark.skipif(not cil.cython_funcs_available, reason='Cython functions are not available')
def test_equivalence_cell_idx_list_cython_python_version(single_link):
    #   D2Q9
    stencil_2d = tuple((x, y) for x, y in product([-1, 0, 1], [-1, 0, 1]))
    #   D3Q19
    stencil_3d = tuple(
        (x, y, z) for x, y, z in product([-1, 0, 1], [-1, 0, 1], [-1, 0, 1]) if abs(x) + abs(y) + abs(z) < 3)

    for dtype in [int, np.int16, np.uint32]:
        fluid_mask = dtype(1)
        mask = dtype(2)
        flag_field_2d = np.ones([15, 16], dtype=dtype) * fluid_mask
        flag_field_3d = np.ones([15, 16, 17], dtype=dtype) * fluid_mask

        flag_field_2d[0, :] = mask
        flag_field_2d[-1, :] = mask
        flag_field_2d[7, 7] = mask

        flag_field_3d[0, :, :] = mask
        flag_field_3d[-1, :, :] = mask
        flag_field_3d[7, 7, 7] = mask

        result_python_2d = cil._create_index_list_python(flag_field_2d, mask, fluid_mask,
                                                         stencil_2d, single_link, False)

        result_python_3d = cil._create_index_list_python(flag_field_3d, mask, fluid_mask,
                                                         stencil_3d, single_link, False)

        result_cython_2d = cil.create_boundary_index_list(flag_field_2d, stencil_2d, mask, fluid_mask, None,
                                                          False, single_link)
        result_cython_3d = cil.create_boundary_index_list(flag_field_3d, stencil_3d, mask, fluid_mask, None,
                                                          False, single_link)

        np.testing.assert_equal(result_python_2d, result_cython_2d)
        np.testing.assert_equal(result_python_3d, result_cython_3d)


@pytest.mark.parametrize('inner_or_boundary', [False, True])
def test_normal_calculation(inner_or_boundary):
    stencil = tuple((x, y) for x, y in product([-1, 0, 1], [-1, 0, 1]))
    domain_size = (32, 32)
    dtype = np.uint32
    fluid_mask = dtype(1)
    mask = dtype(2)
    flag_field = np.ones([domain_size[0], domain_size[1]], dtype=dtype) * fluid_mask

    radius_inner = domain_size[0] // 4
    radius_outer = domain_size[0] // 2
    y_mid = domain_size[1] / 2
    x_mid = domain_size[0] / 2

    for x in range(0, domain_size[0]):
        for y in range(0, domain_size[1]):
            if (y - y_mid) ** 2 + (x - x_mid) ** 2 < radius_inner ** 2:
                flag_field[x, y] = mask
            if (x - x_mid) ** 2 + (y - y_mid) ** 2 > radius_outer ** 2:
                flag_field[x, y] = mask

    args_no_gl = (flag_field, mask, fluid_mask, np.array(stencil, dtype=np.int32), True)
    index_list = cil._create_index_list_python(*args_no_gl, inner_or_boundary=inner_or_boundary, nr_of_ghost_layers=1)

    checkmask = mask if inner_or_boundary else fluid_mask

    for cell in index_list:
        idx = cell[2]
        cell = tuple((cell[0], cell[1]))
        sum_cells = np.zeros(len(cell))
        for dir_idx, direction in enumerate(stencil):
            neighbor_cell = tuple([cell_i + dir_i for cell_i, dir_i in zip(cell, direction)])
            if any(not 0 <= e < upper for e, upper in zip(neighbor_cell, flag_field.shape)):
                continue
            if flag_field[neighbor_cell] & checkmask:
                sum_cells += np.array(direction)

        assert np.argmax(np.inner(sum_cells, stencil)) == idx



import numpy as np
import sympy as sp
import pytest

import pystencils as ps
from pystencils.astnodes import Block, Conditional, SympyAssignment
from pystencils.backends.simd_instruction_sets import get_supported_instruction_sets, get_vector_instruction_set
from pystencils.enums import Target
from pystencils.cpu.vectorization import vec_all, vec_any
from pystencils.node_collection import NodeCollection

supported_instruction_sets = get_supported_instruction_sets() if get_supported_instruction_sets() else []


@pytest.mark.parametrize('instruction_set', supported_instruction_sets)
@pytest.mark.parametrize('dtype', ('float', 'double'))
def test_vec_any(instruction_set, dtype):
    if instruction_set in ['sve', 'rvv']:
        width = 4  # we don't know the actual value
    else:
        width = get_vector_instruction_set(dtype, instruction_set)['width']
    data_arr = np.zeros((4 * width, 4 * width), dtype=np.float64 if dtype == 'double' else np.float32)

    data_arr[3:9, 1:3 * width - 1] = 1.0
    data = ps.fields(f"data: {dtype}[2D]", data=data_arr)

    c = [
        SympyAssignment(sp.Symbol("t1"), vec_any(data.center() > 0.0)),
        Conditional(vec_any(data.center() > 0.0), Block([SympyAssignment(data.center(), 2.0)]))
    ]

    assignmets = NodeCollection(c)
    ast = ps.create_kernel(assignments=assignmets, target=ps.Target.CPU,
                           cpu_vectorize_info={'instruction_set': instruction_set})
    kernel = ast.compile()
    kernel(data=data_arr)
    if instruction_set in ['sve', 'rvv']:
        # we only know that the first value has changed
        np.testing.assert_equal(data_arr[3:9, :3 * width - 1], 2.0)
    else:
        np.testing.assert_equal(data_arr[3:9, :3 * width], 2.0)


@pytest.mark.parametrize('instruction_set', supported_instruction_sets)
@pytest.mark.parametrize('dtype', ('float', 'double'))
def test_vec_all(instruction_set, dtype):
    if instruction_set in ['sve', 'rvv']:
        width = 1000  # we don't know the actual value, need something guaranteed larger than vector
    else:
        width = get_vector_instruction_set(dtype, instruction_set)['width']
    data_arr = np.zeros((4 * width, 4 * width), dtype=np.float64 if dtype == 'double' else np.float32)

    data_arr[3:9, 1:3 * width - 1] = 1.0
    data = ps.fields(f"data: {dtype}[2D]", data=data_arr)

    c = [Conditional(vec_all(data.center() > 0.0), Block([SympyAssignment(data.center(), 2.0)]))]
    assignmets = NodeCollection(c)
    ast = ps.create_kernel(assignmets, target=Target.CPU,
                           cpu_vectorize_info={'instruction_set': instruction_set})
    kernel = ast.compile()
    kernel(data=data_arr)
    if instruction_set in ['sve', 'rvv']:
        # we only know that some values in the middle have been replaced
        assert np.all(data_arr[3:9, :2] <= 1.0)
        assert np.any(data_arr[3:9, 2:] == 2.0)
    else:
        np.testing.assert_equal(data_arr[3:9, :1], 0.0)
        np.testing.assert_equal(data_arr[3:9, 1:width], 1.0)
        np.testing.assert_equal(data_arr[3:9, width:2 * width], 2.0)
        np.testing.assert_equal(data_arr[3:9, 2 * width:3 * width - 1], 1.0)
        np.testing.assert_equal(data_arr[3:9, 3 * width - 1:], 0.0)


@pytest.mark.skipif(not supported_instruction_sets, reason='cannot detect CPU instruction set')
def test_boolean_before_loop():
    t1, t2 = sp.symbols('t1, t2')
    f_arr = np.ones((10, 10))
    g_arr = np.zeros_like(f_arr)
    f, g = ps.fields("f, g : double[2D]", f=f_arr, g=g_arr)

    a = [
        ps.Assignment(t1, t2 > 0),
        ps.Assignment(g[0, 0],
                      sp.Piecewise((f[0, 0], t1), (42, True)))
    ]
    ast = ps.create_kernel(a, cpu_vectorize_info={'instruction_set': supported_instruction_sets[-1]})
    kernel = ast.compile()
    kernel(f=f_arr, g=g_arr, t2=1.0)
    # print(g)
    np.testing.assert_array_equal(g_arr, 1.0)
    kernel(f=f_arr, g=g_arr, t2=-1.0)
    np.testing.assert_array_equal(g_arr, 42.0)


@pytest.mark.parametrize('instruction_set', supported_instruction_sets)
@pytest.mark.parametrize('dtype', ('float32', 'float64'))
def test_vec_maskstore(instruction_set, dtype):
    data_arr = np.zeros((16, 16), dtype=np.float64 if dtype == 'float64' else np.float32)
    data_arr[3:-3, 3:-3] = 1.0
    data = ps.fields(f"data: {dtype}[2D]", data=data_arr)

    c = [Conditional(data.center() < 1.0, Block([SympyAssignment(data.center(), 2.0)]))]

    assignmets = NodeCollection(c)
    config = ps.CreateKernelConfig(cpu_vectorize_info={'instruction_set': instruction_set}, default_number_float=dtype)
    ast = ps.create_kernel(assignmets, config=config)
    print(ps.get_code_str(ast))
    kernel = ast.compile()
    kernel(data=data_arr)
    np.testing.assert_equal(data_arr[:3, :], 2.0)
    np.testing.assert_equal(data_arr[-3:, :], 2.0)
    np.testing.assert_equal(data_arr[:, :3], 2.0)
    np.testing.assert_equal(data_arr[:, -3:], 2.0)
    np.testing.assert_equal(data_arr[3:-3, 3:-3], 1.0)
